"""
    Alsaker.Bridge.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from time import time
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.ATSAS.AutorgRunner import AutorgRunner

VERY_SMALL_VALUE = 1e-10

def compare_bridge_impl(caller, corrected_sd=None):
    from importlib import reload
    import molass_legacy.Alsaker.Bridge
    reload(molass_legacy.Alsaker.Bridge)
    from molass_legacy.Alsaker.Bridge import RgcodeBrige, do_it_in_the_context

    logger = logging.getLogger(__name__)
    print("compare_bridge_impl")
    if corrected_sd is None:
        sd = caller.serial_data
    else:
        sd = corrected_sd
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y

    D[D < VERY_SMALL_VALUE] = VERY_SMALL_VALUE

    compute_all_points = True
    check_peak_tops = False
    check_y3 = False

    def compute_molass_rgs():
        sg_results = []
        t0 = time()
        for j in range(D.shape[1]):
            if j%10 == 0:
                print([j])
            data = np.array([qv, D[:,j], E[:,j]]).T
            try:
                sg = SimpleGuinier(data)
                sq_result = (sg.Rg, sg.basic_quality)
            except:
                sq_result = (np.nan, np.nan)
            sg_results.append(sq_result)
        sg_array = np.array(sg_results)
        print("SG time elapsed: %f" % (time() - t0))
        return sg_array

    def compute_atsas_rgs():
        at = AutorgRunner()
        at_results = []
        t0 = time()
        for j in range(D.shape[1]):
            if j%10 == 0:
                print([j])
            data = np.array([qv, D[:,j], E[:,j]]).T
            orig_result, _ = at.run_from_array(data)
            at_results.append((orig_result.Rg, orig_result.Rg_stdev))
        at_array = np.array(at_results)
        print("AT time elapsed: %f" % (time() - t0))
        return at_array

    def compute_peaktop_rgs(rb):
        data_list = []
        if check_y3:
            y3_list = []

        for rec in xr_curve.peak_info:
            j = rec[1]
            print([j])
            dv = D[:,j]
            ev = E[:,j]
            data = np.array([qv, dv, ev]).T
            data_list.append(data)
            where_negative = np.where(dv <= 0)[0]
            if len(where_negative) == 0:
                starting_value = None
            else:
                starting_value = where_negative[-1] + 1
            print([j], "starting_value", starting_value)
            if check_y3:
                print([j], "np.where(np.isnan(data[:,1])", np.where(np.isnan(data[:,1]))[0])
                y3 = rb.estimate_Rg(data, 1, starting_value=starting_value, return_y3=True)
                print([j], "np.where(np.isnan(y3)", np.where(np.isnan(y3))[0])
                y3_list.append(y3)
            else:
                try:
                    output = rb.estimate_Rg(data, 1, starting_value=starting_value)
                    print([j], starting_value, output)
                except:
                    log_exception(logger, "estimate_Rg failed")

        with plt.Dp():
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
            ax1.plot(x, y)
            for rec in xr_curve.peak_info:
                j = rec[1]
                ax1.plot(x[j], y[j], "o", color="red")

            for data in data_list:
                qv_, dv, ev = data.T
                ax2.plot(qv_, dv)
            if check_y3:
                for y3 in y3_list:
                    ax3.plot(y3)
            fig.tight_layout()
            plt.show()

    def compute_alsaker_rgs(rb):
        t0 = time()
        al_results = []
        for j in range(D.shape[1]):
            if j%10 == 0:
                print([j])
            data = np.array([qv, D[:,j], E[:,j]]).T
            try:
                output = rb.estimate_Rg(data, 1)
            except:
                output = (np.nan, np.nan)
            al_results.append((output[0], output[1]))
        al_array = np.array(al_results)
        print("AL time elapsed: %f" % (time() - t0))
        return al_array

    if compute_all_points:
        sg_array = compute_molass_rgs()
        at_array = compute_atsas_rgs()

    def closure():
        rb = RgcodeBrige()

        if check_peak_tops:
            compute_peaktop_rgs(rb)

        if compute_all_points:
            al_array = compute_alsaker_rgs(rb)
            return sg_array, at_array, al_array

    return do_it_in_the_context(closure)



# coding: utf-8
"""

    ファイル名：   GuinierStartSelector.py

    処理内容：

        Guinier区間開始点の選択画面

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""

import numpy                as np
from scipy                  import stats
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from matplotlib.patches     import Polygon
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar

GUNIER_PLOT_SIZE    = 100
GUNIER_START_MAX    = 60
GUNIER_SIZE_EXCEPT  = 20

class GuinierStartSelector( Dialog ):
    def __init__( self, parent, serial_data, init_ranges ):

        self.created_parent = False

        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = 'Guinier interval start point selection'
        self.serial_data    = serial_data
        self.init_ranges    = init_ranges
        self.num_peaks      = len(init_ranges)
        self.gs         = None
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )

        self.peak_no    = self.parent.guinier_fig_peak_no.get()
        self.set_plot_data()

    def set_plot_data( self ):

        self.one_peak = self.init_ranges[self.peak_no][1][1]
        # print( self.init_ranges, one_peak )
        one_peak_data = self.serial_data.intensity_array[self.one_peak, :, : ]
        x_  = one_peak_data[0:GUNIER_PLOT_SIZE, 0]
        y_  = one_peak_data[0:GUNIER_PLOT_SIZE, 1]
        e_  = one_peak_data[0:GUNIER_PLOT_SIZE, 2]
        positive = np.logical_and( y_ > 0, e_ > 0 )

        self.x      = x_[positive]
        self.y      = y_[positive]
        self.x2     = self.x**2
        self.log_y  = np.log( y_[positive] )

    def show( self ):
        self.figsize    = ( 16, 8 )
        self.message    = None
        self.button_labels  = [ "OK", "Cancel" ]
        self.toolbar    = True

        Dialog.__init__( self, self.parent, self.title_ )
        # TODO: self.resizable(width=False, height=False)

    def destroy_parent( self ):
        if self.created_parent:
            self.parent.destroy()

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        if self.message is not None:
            self.msg = Tk.Label( body_frame, text=self.message, bg='white' )
            self.msg.pack( fill=Tk.BOTH, expand=1, pady=20 )
            # msg.insert( Tk.INSERT, self.message )
            # msg.config( state=Tk.DISABLED )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        frame_ = Tk.Frame( body_frame )
        frame_.pack( fill=Tk.X, padx=200 )

        frame_left = Tk.Frame( frame_ )
        frame_left.pack( side=Tk.LEFT )

        label_ = Tk.Label( frame_left, text="Peak selection for Guinier plot: " )
        label_.pack( side=Tk.LEFT  )

        self.guinier_fig_peak_no = Tk.IntVar()
        self.guinier_fig_peak_no.set( self.parent.guinier_fig_peak_no.get() )
        entry_ = Tk.Spinbox( frame_left, textvariable=self.guinier_fig_peak_no,
                                            from_=0, to=self.num_peaks-1, increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        entry_.pack( side=Tk.LEFT  )
        self.guinier_fig_peak_no.trace( "w", self.guinier_fig_peak_no_tracer )

        frame_right = Tk.Frame( frame_ )
        frame_right.pack( side=Tk.RIGHT )

        label_ = Tk.Label( frame_right, text="Guinier interval start point: " )
        label_.pack( side=Tk.LEFT  )

        self.guinier_start_point = Tk.IntVar()
        self.guinier_start_point.set( self.parent.guinier_start_point.get() )
        entry_ = Tk.Spinbox( frame_right, textvariable=self.guinier_start_point,
                                            from_=0, to=50, increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        entry_.pack( side=Tk.LEFT  )
        self.guinier_start_point.trace( "w", self.guinier_start_point_tracer )

        figsize_ = ( 18, 9 ) if self.figsize is None else self.figsize

        fig = plt.figure( figsize=figsize_ )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        self.draw_func( fig )
        self.parent.config( cursor='' )

        if self.toolbar:
            self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
            self.toolbar.update()

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text=self.button_labels[0], width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text=self.button_labels[1], width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def ok( self ):
        self.destroy_parent()
        self.applied  = True
        self.parent.guinier_start_point.set( self.guinier_start_point.get() )
        self.parent.guinier_fig_peak_no.set( self.guinier_fig_peak_no.get() )
        self.destroy()

    def cancel( self ):
        self.destroy_parent()
        self.applied  = False
        self.destroy()

    def draw_func( self, fig, clear=False, start=None ):
        if start is None:
            start   = self.guinier_start_point.get()

        if self.gs is None:
            self.fig = fig
            self.gs  = gridspec.GridSpec( 4, 2 )
            self.ax1 = fig.add_subplot( self.gs[0:3, 0] )
            self.ax2 = fig.add_subplot( self.gs[0:3, 1] )
            self.ax3 = fig.add_subplot( self.gs[3, 1] )
            self.ax4 = fig.add_subplot( self.gs[3, 0] )
            self.axes = [ self.ax1, self.ax2, self.ax3, self.ax4 ]

        ax1 = self.ax1
        ax2 = self.ax2
        ax3 = self.ax3
        ax4 = self.ax4

        if clear:
            for ax in self.axes:
                ax.cla()

        self.ax1.set_title( 'Guinier plot at elution No. %d' % self.one_peak )
        self.ax2.set_title( 'Zoomed plot of the left fig.' )
        self.ax3.set_title( 'Residuals in the Guinier interval' )
        self.ax4.set_title( 'Xray elution curve' )

        ax1.plot( self.x2, self.log_y, 'o', markersize=3 )
        ax2.plot( self.x2[0:GUNIER_START_MAX], self.log_y[0:GUNIER_START_MAX], 'o',  markersize=3 )

        for ax in self.axes[0:3]:
            ax.set_xlabel( 'Q²' )
            ax.set_ylabel( 'ln(I)' )

        ax4.set_xlabel( 'elution №' )
        ax4.set_ylabel( 'Intensity' )
        ivector = self.serial_data.ivector
        ax4.plot( ivector, color='orange' )
        ax4.plot( [ self.one_peak, self.one_peak ], [ 0, ivector[self.one_peak] ] )
        py = self.y[self.guinier_start_point.get()]
        ax4.plot( self.one_peak, py, 'o', color='red', markersize=5 )

        ax1.set_xlim( ax1.get_xlim() )
        ax1.set_ylim( ax1.get_ylim()  )

        xmin, xmax = ax2.get_xlim()
        ymin, ymax = ax2.get_ylim()

        ax2.set_xlim( xmin, xmax )
        ax2.set_ylim( ymin, ymax )
        xoffset = ( xmax + xmin ) * 0.05
        yoffset = ( ymax - ymin ) * 0.05
        for i in [ 10, 20, 30 ]:
            x = self.x2[i]
            ax2.plot( [x, x], [ymin, ymax], '-', color='gray', alpha=0.5 )
            ax2.annotate( '%d-th Q' % i, xy=(x, ymin), alpha=0.5,
                            xytext=( x + xoffset, ymin + yoffset ),
                            arrowprops=dict( headwidth=3, width=0.5, color='black', alpha=0.5 ),
                            )

        ax1.plot( [ xmin, xmax, xmax, xmin, xmin ], [ ymin, ymin, ymax, ymax, ymin ], ':', color='gray', alpha=0.5 )

        self.draw_elution_curve_area()
        self.draw_guinier_interval( start )

        fig.tight_layout()

    def guinier_start_point_tracer( self, *args ):
        try:
            start   = self.guinier_start_point.get()
        except:
            return

        self.draw_func( self.fig, clear=True, start=start )
        self.mpl_canvas.draw()

    def draw_elution_curve_area( self ):
        xmin, xmax = self.ax1.get_xlim()
        ymin, ymax = self.ax1.get_ylim()

        start   = self.serial_data.xray_slice.start
        stop    = self.serial_data.xray_slice.stop + 1  # include stop
        slice_  = slice( start, stop )
        x2_     = self.x2[slice_]
        log_y_  = self.log_y[slice_]
        verts = [ (x2_[0], ymin) ] + list(zip(x2_, log_y_)) + [(x2_[-1], ymin)]

        for ax in self.axes[0:2]:
            # you must give a Polygon separately for each axis
            poly = Polygon(verts, facecolor='orange', alpha=0.2 )
            ax.add_patch(poly)

        xoffset = ( xmax + xmin ) * 0.1

        tx  = np.average( x2_ )
        ty  = ymin*0.9 + ymax*0.1
        self.ax1.annotate( 'Range used for elution curve averaging',
                        xy=(tx, ty), alpha=0.5, va='center',
                        xytext=( tx + xoffset, ty ), 
                        arrowprops=dict( headwidth=3, width=0.5, color='black', alpha=0.5 ),
                        )

    def draw_guinier_interval( self, start ):
        stop = self.compute_guinier_stop( start )
        # print( 'qrg_limit=', qrg_limit )
        xr  = self.x2[start:stop]
        yr  = self.log_y[start:stop]
        slope, intercept, r_value, p_value, std_err = stats.linregress( xr, yr )

        x   = self.x2[ [start, stop] ]
        y   = slope*x + intercept
        for ax in [self.ax1, self.ax2]:
            ax.plot( x, y, color='red' )
            ax.plot( xr[0], yr[0], 'o', color='red', markersize=5 )

        self.ax3.set_xlim( self.ax2.get_xlim() )

        self.ax3.plot( x, [ 0, 0 ], color='red' )
        yr_ = yr - ( slope*xr + intercept )
        self.ax3.plot( xr, yr_ )
        self.ax3.plot( xr[0], yr_[0], 'o', color='red', markersize=5 )

    def compute_guinier_stop( self, start ):
        stop_range = slice( start+1, None )
        x2_ = self.x2[stop_range]
        log_y_ = self.log_y[stop_range]
        slope   =  (log_y_ - self.log_y[start]) /( x2_ - self.x2[start] )
        rg = np.sqrt( -3*slope )
        qrg = self.x[stop_range] * rg
        try:
            qrg_sizet = np.where( qrg < 1.3 )[0][-1]
        except:
            qrg_sizet = GUNIER_SIZE_EXCEPT
        return start + qrg_sizet + 1

    def guinier_fig_peak_no_tracer( self, *args ):
        try:
            self.peak_no    = self.guinier_fig_peak_no.get()
        except:
            return
        self.set_plot_data()
        self.draw_func( self.fig, clear=True )
        self.mpl_canvas.draw()



# coding: utf-8
"""

    MplCanvas.py

    to avoid
    invalid command name "1849982244872filter_destroy"
    from DerbugPlot

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
import tkinter as Tk
import matplotlib
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg

class MplCanvas(Tk.Frame):
    def __init__(self, parent, *args):
        self.parent = parent
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'
        Tk.Frame.__init__(self, parent, *args)

    def attach_figure(self, fig):
        print( fig.get_size_inches() )
        self.mpl_canvas = FigureCanvasTkAgg( fig, self )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

    def detach_figure(self):
        pass

    def destroy(self):
        self.detach_figure()
        super(Tk.Frame,self).destroy()

    def show(self):
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        print( "MplCanvas: reqsize", w, h )

        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()

        # TODO: resize appropriately
        # https://ja.osdn.net/projects/pylaf/scm/hg/pylaf/blobs/tip/src/pylafiii/mplext.py
        print(self.mpl_canvas_widget.winfo_geometry())



# coding: utf-8
"""

    LinkedToplevel.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""

from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.TkUtils import adjusted_geometry, split_geometry, join_geometry

class LinkedToplevel(Tk.Toplevel):
    def __init__(self, root, body=None):
        self.root = root
        self.built = False
        self.body = body
        Tk.Toplevel.__init__(self)

    def show(self):
        if not self.built:
            if self.body is not None:
                self.body(self, self.body_frame, self.navi_frame)
                self.built = True
        self.deiconify()

    def link(self, prev=None, next=None, body_kwargs=dict(width=400, height=300), navi_kwargs={}):
        self._prev = prev
        self._next = next

        self.body_frame = Tk.Frame(self, **body_kwargs)
        self.body_frame.pack(fill=Tk.BOTH, expand=1)
        self.navi_frame = Tk.Frame(self, **navi_kwargs)
        self.navi_frame.pack(fill=Tk.BOTH, expand=0)

        if self._prev is None:
            button = Tk.Button(self.navi_frame, text="◀ Exit", command=self.exit)
        else:
            button = Tk.Button(self.navi_frame, text="◀ Back", command=self._show_prev)
        button.pack(side=Tk.LEFT, padx=5, pady=5)
        self._prev_button = button

        if self._next is None:
            button = None
        else:
            button = Tk.Button(self.navi_frame, text="▶ Next", command=self._show_next)
            button.pack(side=Tk.RIGHT, padx=5)
        self._next_button = button
        self.update_geometry()
        self.restore_geometry()
        self.withdraw()

    def update_geometry(self):
        self.update()
        self.geometry_save = adjusted_geometry(self.geometry())

    def restore_geometry(self, invoked_geometry=None):
        if invoked_geometry is None:
            geometry = self.geometry_save
        else:
            _, _, x, y = split_geometry(invoked_geometry)
            w, h, _, _ = split_geometry(self.geometry_save)
            geometry = join_geometry(w, h, x, y)

        self.geometry(geometry)

    def rename_navi_buttons(self, names):
        for name, button in zip(names, [self._prev_button, self._next_button]):
            if name is None:
                continue
            button.config(text=name)

    def exit(self):
        toplevels = []
        next_ = self._next
        while True:
            if next_ is None:
                break
            toplevels.append(next_)
            next_ = next_._next

        for w in toplevels:
            w.destroy()

        self.destroy()
        self.root.quit()

    def _show_prev(self):
        geometry = self.geometry()
        self.withdraw()
        geometry = self.geometry()
        self._prev.show()
        self._prev.restore_geometry(invoked_geometry=geometry)

    def _show_next(self):
        geometry = self.geometry()
        self.withdraw()
        self._next.show()
        self._next.restore_geometry(invoked_geometry=geometry)



"""
    Optimizer.GuinierDeviationTester.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def test_update_guinier_region_impl(js_canvas):
    print("test_update_guinier_region_impl")
    optimizer = js_canvas.fullopt
    params = js_canvas.get_current_params()
    optimizer.update_guinier_region(params=params, debug=True)

def show_guinier_deviation(js_canvas):
    from ..Optimizer.FvScoreConverter import convert_score
    print("show_guinier_deviation")
    optimizer = js_canvas.fullopt
    params = js_canvas.get_current_params()
    lrf_info = optimizer.objective_func(params, return_lrf_info=True)
    Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices
    rg_params = optimizer.separate_params[2]
    deviation = optimizer.get_guinier_deviation(Pxr, rg_params, debug=True)
    print("deviation=", deviation, convert_score(deviation))



"""
    UnknownConc.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from bisect import bisect_right
from molass_legacy.Peaks.ElutionModels import egh
import seaborn
seaborn.set()
import molass_legacy.KekLib.DebugPlot as plt

def demo1_():
    qv = np.linspace(0.005, 0.4, 400)
    jv = np.arange(300)
    rg = 35
    h = 1
    mu = 150
    sigma = 20
    tau = 0
    a = 0
    K = 0.5
    error = 0.05
    demo_impl(qv, jv, rg, h, mu, sigma, tau, a, K, error)

def demo1():
    qv = np.linspace(0.005, 0.4, 400)
    jv = np.arange(300)
    rg = 30
    h = 1
    mu = 211
    sigma = 23.6
    tau = -29.5
    a = -0.49
    K = 0.5
    error = 0.02
    demo_impl(qv, jv, rg, h, mu, sigma, tau, a, K, error)

def demo_impl(qv, jv, rg, h, mu, sigma, tau, a, K, error):
    from importlib import reload
    import SimTools.HardSphere
    reload(SimTools.HardSphere)
    from SimTools.HardSphere import get_model_data

    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, a=a, K=K, error=0)

    aq, bq = P.T

    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, a=a, K=K, error=error)

    i = bisect_right(qv, 0.02)

    cy = D[i,:]
    C_ = np.array([cy, cy**2])
    C_inv = np.linalg.pinv(C_)
    P_ = D @ C_inv

    """
    Iqc1 = aq*c1 + bq*c1**2
    Iqc2 = aq*c2 + bq*c2**2
        ...
        ...
    Iqcj = aq*cj + bq*cj**2
        ...
    """

    aq_, bq_ = P_.T

    def objective(p):
        h, m, s, t, a, b = p
        cy_ = egh(jv, h, m, s, t)
        return np.linalg.norm(a*cy_ + b*cy_**2 - cy)

    j = np.argmax(cy)
    h = cy[j]
    s = 10
    t = 0

    init_params = h, j, s, t, aq_[i], bq_[i]
    ret = minimize(objective, init_params)
    print("ret.fun=", ret.fun)

    h, m, s, t, a, b = ret.x
    cy_ = egh(jv, h, m, s, t)

    Cs = np.array([cy_, cy_**2])
    Csinv = np.linalg.pinv(Cs)
    Ps = D @ Csinv
    aqs, bqs = Ps.T

    plot_demo_result(qv, jv, cy, cy_, aq, bq, aq_, bq_, aqs, bqs, rg, error, D, j)

def plot_demo_result(qv, jv, cy, cy_, aq, bq, aq_, bq_, aqs, bqs, rg, error, D, j):
    scale = np.max(aq)/np.max(aqs)
    print("scale=", scale)
    aqs_ = scale*aqs
    bqs_ = scale**2*bqs

    ig = bisect_right(qv, 1.8/rg)
    gslice = slice(0,ig)
    gqv = qv[gslice]
    gqv2 = gqv**2

    gy = D[gslice,j]

    with plt.Dp():
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,5))
        fig.suptitle("True A(q), B(q) retrieval in Hard Sphere Model with Error=%g" % error, fontsize=20)
        ax1.set_title("Apparent Estimation", fontsize=16)
        ax2.set_title("Concetration Correction", fontsize=16)
        ax3.set_title("Corrected Estimation", fontsize=16)
        ax4.set_title("Guinier Comparison", fontsize=16)

        ax1.plot(qv, aq, label="True A(q)")
        ax1.plot(qv, bq, label="True B(q)", color="pink")
        ax1.plot(qv, aq_, ":", label="A(q) directly from measured data", color="C1", lw=2)
        ax1.plot(qv, bq_, ":", label="B(q) directly from measured data", color="cyan", lw=2)
        ax2.plot(jv, cy, label="apparent concentration", color="gray", alpha=0.5)
        ax2.plot(jv, cy_, ":", label="corrected concentration", color="C2", lw=2)
        ax3.plot(qv, aq, label="True A(q)")
        ax3.plot(qv, bq, label="True B(q)", color="pink")
        ax3.plot(qv, aqs_, ":", label="retrieved A(q)", color="C2", lw=2)
        ax3.plot(qv, bqs_, ":", label="retrieved B(q)", color="C3", lw=2)

        ax4.plot(gqv2, np.log(gy), label="apparent data", color="gray", alpha=0.5)
        ax4.plot(gqv2, np.log(aq[gslice]), label="true")
        ax4.plot(gqv2, np.log(aq_[gslice]), ":", label="naiive A(q)", color="C1", lw=2)
        ax4.plot(gqv2, np.log(aqs_[gslice]), ":", label="retrieved A(q)", color="C2", lw=2)

        q = 1.3/rg
        q2 = q**2
        ymin, ymax = ax4.get_ylim()
        ax4.set_ylim(ymin,ymax)
        ax4.plot([q2, q2], [ymin, ymax], color="yellow", label="$qR_g=1.3$")

        for ax in ax1, ax2, ax3, ax4:
            ax.legend()

        fig.tight_layout()
        plt.show()

def realistic_demo_impl(qv, jv, rg, h, mu, sigma, tau, a, K, error):
    from importlib import reload
    import SimTools.HardSphere
    reload(SimTools.HardSphere)
    from SimTools.HardSphere import get_model_data

    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, a=a, K=K, error=0)

    aq, bq = P.T

    D, E, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, a=a, K=K, error=error, return_error_matrix=True)

    i = bisect_right(qv, 0.02)

    cy = D[i,:]
    C_ = np.array([cy, cy**2])
    C_inv = np.linalg.pinv(C_)
    P_ = D @ C_inv

    """
    Iqc1 = aq*c1 + bq*c1**2
    Iqc2 = aq*c2 + bq*c2**2
        ...
        ...
    Iqcj = aq*cj + bq*cj**2
        ...
    """

    aq_, bq_ = P_.T

    def objective(p):
        h, m, s, t, a, b = p
        cy_ = egh(jv, h, m, s, t)
        return np.linalg.norm(a*cy_ + b*cy_**2 - cy)

    j = np.argmax(cy)
    h = cy[j]
    s = 10
    t = 0

    init_params = h, j, s, t, aq_[i], bq_[i]
    ret = minimize(objective, init_params)
    print("ret.fun=", ret.fun)

    h, m, s, t, a, b = ret.x
    cy_ = egh(jv, h, m, s, t)

    Cs = np.array([cy_, cy_**2])
    Csinv = np.linalg.pinv(Cs)
    Ps = D @ Csinv
    aqs, bqs = Ps.T

    plot_demo_result(qv, jv, cy, cy_, aq, bq, aq_, bq_, aqs, bqs, rg, error, D, j)



# coding: utf-8
"""
    ConcNormalizerDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import re
from difflib import SequenceMatcher
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox
from SerialDataUtils import get_xray_files
from ConcNormalizer import find_conc_file, normalize_impl

class ConcNormalizerDialog(Dialog):
    def __init__(self, parent):
        self.busy = False
        Dialog.__init__(self, parent, "Concentration Normalizer", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        folder_entry_width = 60
        num_files_lable_width = 6

        grid_row = -1

        grid_row += 1
        label = Tk.Label(iframe, text="Folder Path")
        label.grid(row=grid_row, column=1)
        label = Tk.Label(iframe, text="File Name Pattern")
        label.grid(row=grid_row, column=2, padx=10)
        label = Tk.Label(iframe, text="Files")
        label.grid(row=grid_row, column=3)
        label = Tk.Label(iframe, text="Conc File")
        label.grid(row=grid_row, column=4)


        grid_row += 1
        label = Tk.Label(iframe, text="Input")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        self.in_folder = Tk.StringVar()
        self.in_files = []
        self.name_pattern = Tk.StringVar()
        self.num_files = Tk.IntVar()
        in_folder_entry = FolderEntry(iframe, textvariable=self.in_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_in_folder)
        in_folder_entry.grid(row=grid_row, column=1)
        in_folder_entry.bind('<Leave>', lambda *args: self.on_entry_in_folder())

        label = Tk.Label(iframe, textvariable=self.name_pattern)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files, width=num_files_lable_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        self.conc_file = Tk.StringVar()
        conc_file_label = Tk.Label(iframe, textvariable=self.conc_file, width=12)
        conc_file_label.grid(row=grid_row, column=4)

        grid_row += 1
        label = Tk.Label(iframe, text="Output")
        label.grid(row=grid_row, column=0, sticky=Tk.W)
        self.out_folder = Tk.StringVar()
        out_folder_entry = FolderEntry(iframe, textvariable=self.out_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_out_folder)
        out_folder_entry.grid(row=grid_row, column=1)
        out_folder_entry.bind('<Leave>', lambda *args: self.on_entry_out_folder())

        self.name_pattern_o = Tk.StringVar()
        entry = Tk.Entry(iframe, textvariable=self.name_pattern_o, width=24)
        entry.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        self.num_files_o = Tk.IntVar()
        num_files_lable = Tk.Label(iframe, textvariable=self.num_files_o, width=num_files_lable_width, justify=Tk.RIGHT)
        num_files_lable.grid(row=grid_row, column=3)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Progress")
        label.grid(row=grid_row, column=0)
        self.mpb = ttk.Progressbar(iframe,orient ="horizontal", length=700, mode="determinate")
        self.mpb.grid(row=grid_row, column=1, columnspan=4, padx=5)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        self.normalize_btn = Tk.Button(box, text="Normalize", width=10, command=self.normalize, state=Tk.DISABLED)
        self.normalize_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show()

    def on_entry_in_folder(self):
        in_folder = self.in_folder.get()
        files = get_xray_files(in_folder)
        self.in_files.clear()
        self.in_files += files
        if len(files) > 0:
            dir_, file = os.path.split(files[0])
            self.name_pattern.set(file)
        self.num_files.set( len(files) )
        file = find_conc_file(in_folder)
        self.conc_file.set(file)

    def on_entry_out_folder(self):
        name_pattern_o = self.name_pattern.get().replace('.dat', '_cn.dat')
        if is_empty_val(name_pattern_o):
            return

        self.name_pattern_o.set(name_pattern_o)

        out_folder = self.out_folder.get()
        if is_empty_val(out_folder):
            return

        self.normalize_btn.config(state=Tk.NORMAL)

    def normalize(self):
        num_files = len(self.in_files)

        self.busy = True

        conc_file = os.path.join(self.in_folder.get(), self.conc_file.get())

        name_pattern = self.name_pattern_o.get()
        sub_re = re.compile(r'^(.+_\d+)(\D.*)(\.dat)')
        m = sub_re.match(name_pattern)
        assert m
        # print(m.group(1), m.group(2))
        postfix = m.group(2)

        def name_changer(in_file):
            return re.sub(sub_re, lambda m: m.group(1) + postfix + m.group(3), in_file)

        normalize_impl(self.in_files, conc_file, self.out_folder.get(), name_changer, dialog=self)
        self.busy = False

    def is_busy(self):
        return self.busy



# coding: utf-8
"""
    SecTheory.RealisticProof.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from .SecCF import simple_phi
from .SecPDF import FftInvPdf

def single_pore_cf(w, rg, rp, x0, nperm, tperm, me, mp):
    if rg > rp:
        np_ = 0
        tp_ = 0
    else:
        rho = rg/rp
        np_ = nperm*(1 - rho)**me
        tp_ = tperm*(1 - rho)**mp
    time_shift = np.exp(-2*np.pi*1j*x0*w)
    return time_shift*simple_phi(w, np_, tp_)

class RealisticCf:
    def __init__(self, n_species, model_cf):
        self.n_species = n_species
        self.model_cf = model_cf

    def split_params(self, flat_params):
        n = self.n_species*2
        rg_params = flat_params[0:n].reshape((self.n_species,2))
        rp_params = flat_params[n:]
        return rg_params, rp_params

    def __call__(self, w, params):
        rg_params, rp_params = self.split_params(params)
        z = 0
        for rg, weight in rg_params:
            z += weight*self.model_cf(w, rg, *rp_params)
        return z

    def objective_func(self, w, params, cft):
        rg_params, rp_params = self.split_params(params)
        z = 0
        penalty = 0
        for rg, weight in rg_params:
            z += weight*self.model_cf(w, rg, *rp_params)
            penalty += min(0, rg)**2 + max(rg - 200, 0)**2          # rg
            penalty += max(rp_params[0] - 1000, 0)**2               # rp
            x0 = rp_params[1]
            penalty += min(x0, -100)**2 + max(100 - x0, 0)**2       # x0
            tperm = rp_params[3]
            penalty += min(tperm, 0)**2 + max(1 - tperm, 0)**2      # tperm
            for v in rp_params[-2:]:
                penalty += min(0, v)**2 + max(v - 3, 0)**2          # me, mp
            penalty += min(0, weight)**2
        return np.sum(np.abs(z - cft)) + penalty*1e8

    def print_params(self, params):
        rg_params, rp_params = self.split_params(params)
        for k, (rg, weight) in enumerate(rg_params):
            print([k], "%.3g, %.3g" % (rg, weight))
        for name, value in zip(["rp", "x0", "nperm", "tperm", "me", "mp"], rp_params):
            print("%s = %.3g" % (name, value))

def fitting_demo(in_folder, niter=1):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder
    from SecSaxs.DataTreatment import DataTreatment
    from .RealisticSolver import RealisticCfDomain

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()

    fig, ax = plt.subplots()

    x = ecurve.x
    y = ecurve.y

    real_cf = RealisticCf(3, single_pore_cf)
    domain = RealisticCfDomain(x, y, real_cf)
    area = domain.area

    rg_params = np.array([(rg, area*w) for rg, w in zip((70, 60, 50), (0.1, 0.8, 0.1))]).flatten()
    #            rp, nperm, tperm, me, mp
    rp_params = np.array([70, 2000, 1, 1, 1])
    init_params = np.concatenate([rg_params, rp_params])

    if False:
        init_params= np.array([
            9.00000000e+01, -3.03669344e-01,
            6.96923790e+01,  6.50125052e-02,
            6.90057185e+01,  6.46371911e-01,
            7.05385931e+01,  2.64632266e+03,   5.05682855e-01,  3.45833675e-01,  4.90885787e-02])

    for k in range(niter):
        opt_result = domain.fit(init_params=init_params)
        init_params = opt_result.x

    real_cf.print_params(init_params)

    spline = domain.get_spline(init_params)

    ax.plot(x, y, label="data")
    ax.plot(x, spline(x), label="model")
    for k, spline in enumerate(domain.get_component_splines(init_params)):
        ax.plot(x, spline(x), label="component=%d" % k)

    ax.legend()

    fig.tight_layout()
    plt.show()



# coding: utf-8
"""

    PairedRangeLogger.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""

PAIRED_RANGES_LOG_HEADER  = "--- paired_ranges log begin ---"
PAIRED_RANGES_LOG_TRAILER = "--- paired_ranges log end ---"

def log_paired_ranges(logger, paired_ranges):
    # print('log_paied_ranges: ', paired_ranges)

    used_elution_model = None
    model_name = None

    logger.info(PAIRED_RANGES_LOG_HEADER)

    for k, range_ in enumerate(paired_ranges):
        range_str = range_.get_log_str()

        if used_elution_model is None:
            used_elution_model = range_.elm_recs is not None

        if used_elution_model:
            if model_name is None:
                model_name = range_.elm_recs[0][1].get_model_name()
            logger.info("%d-th %s has been decomposed using %s as follows." % (k, range_str, model_name))
            for rec in range_.elm_recs:
                # print('rec=', rec)
                kno = rec[0]        # rec.kno?
                evaluator = rec[1]  # rec.evaluator?
                logger.info("%d-th element: %s" % (kno, evaluator.get_all_params_string()) )
        else:
            logger.info("%d-th %s has no decomposition info" % (k, range_str))

    logger.info(PAIRED_RANGES_LOG_TRAILER)



"""
    DefaultNumPeaks.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Optimizer.OptConstants import MAX_NUM_PEAKS

def get_default_num_peaks(sd):
    xr_curve = sd.get_xray_curve()
    uv_curve = sd.get_uv_curve()
    try:
        xr_num_peaks = len(xr_curve.peak_info)
        uv_num_peaks = len(uv_curve.peak_info)
        if xr_num_peaks == uv_num_peaks:
            num_peaks = xr_num_peaks + 2
            plus = 1
            look_curve = xr_curve
        else:
            num_peaks = max(xr_num_peaks, uv_num_peaks) + 1
            plus = 0
            if xr_num_peaks > uv_num_peaks:
                look_curve = xr_curve
            else:
                look_curve = uv_curve
        ret_num_peaks = min(MAX_NUM_PEAKS, num_peaks)
    except:
        # AttributeError: 'NoneType' object has no attribute 'get_xray_curve'
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "get_default_num_peaks failed: ", n=5)
        raise RuntimeError

    hs = []
    for epeak in look_curve.get_emg_peaks():
        h = epeak.get_params()[0]
        hs.append(h)
    i = np.argmax(hs)
    if plus > 0:
        n = i + plus
    else:
        n = None
    return ret_num_peaks, n



# coding: utf-8
"""
    KnownPaths.py

    excerpts from https://gist.github.com/mkropat/7550097
"""

import ctypes, sys
from ctypes import windll, wintypes
from uuid import UUID

class GUID(ctypes.Structure):   # [1]
    _fields_ = [
        ("Data1", wintypes.DWORD),
        ("Data2", wintypes.WORD),
        ("Data3", wintypes.WORD),
        ("Data4", wintypes.BYTE * 8)
    ] 

    def __init__(self, uuid_):
        ctypes.Structure.__init__(self)
        self.Data1, self.Data2, self.Data3, self.Data4[0], self.Data4[1], rest = uuid_.fields
        for i in range(2, 8):
            self.Data4[i] = rest>>(8 - i - 1)*8 & 0xff

class UserHandle:   # [3]
    current = wintypes.HANDLE(0)
    common  = wintypes.HANDLE(-1)

_CoTaskMemFree = windll.ole32.CoTaskMemFree     # [4]
_CoTaskMemFree.restype= None
_CoTaskMemFree.argtypes = [ctypes.c_void_p]

_SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath     # [5] [3]
_SHGetKnownFolderPath.argtypes = [
    ctypes.POINTER(GUID), wintypes.DWORD, wintypes.HANDLE, ctypes.POINTER(ctypes.c_wchar_p)
] 

class PathNotFoundException(Exception): pass

def get_path(folderid, user_handle=UserHandle.common):
    fid = GUID(folderid) 
    pPath = ctypes.c_wchar_p()
    S_OK = 0
    if _SHGetKnownFolderPath(ctypes.byref(fid), 0, user_handle, ctypes.byref(pPath)) != S_OK:
        raise PathNotFoundException()
    path = pPath.value
    _CoTaskMemFree(pPath)
    return path

def get_downloads_folder():
    Downloads = UUID('{374DE290-123F-4565-9164-39C4925E467B}')
    return str(get_path(Downloads))



"""
    Sigmoid.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from scipy.optimize import curve_fit, minimize
from bisect import bisect_right

FC_CERTAINTY_SCALE = 7      # np.exp(-0.1 * 7) == 0.5

def sigmoid(x, L ,x0, k, b):
    return L/(1 + np.exp(-k*(x-x0))) + b

def sigmoid_inv(y, L ,x0, k, b):
    """
        y = L/(1 + np.exp(-k*(x-x0))) + b
        y - b = L/
        (1 + np.exp(-k*(x-x0))) = L/(y - b)
        np.exp(-k*(x-x0)) = L/(y - b) - 1
        -k*(x-x0) = np.log(L/(y - b) - 1)
        x = -np.log(L/(y - b) - 1)/k + x0
    """
    return -np.log(L/(y - b) - 1)/k + x0

def bent_sigmoid(x, L , x0, k, b, s1, s2):
    y = sigmoid(x, L, x0, k, b)
    x_ = x - x0
    return y + np.hstack([s1*x_[x_ < 0], s2*x_[x_ >= 0]])

def bent_sigmoid_inv(y, L ,x0, k, b, s1, s2, debug=False):
    assert np.isscalar(y)

    nx = np.arange(x0-10, x0+11)
    ny = bent_sigmoid(nx, L, x0, k, b, s1, s2)
    x_ = nx - x0
    y_ = ny - np.hstack([s1*x_[x_ < 0], s2*x_[x_ >= 0]])
    if L > 0:
        i = bisect_right(y_, y)
    else:
        i = bisect_right(-y_, -y)
    ret_x = x0 + x_[i]

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("bent_sigmoid_inv debug")
            ax.plot(nx, ny)
            ax.plot(ret_x, y, "o")
            fig.tight_layout()
            plt.show()

    return ret_x

def guess_x0(x, y, debug=False):
    if debug:
        from importlib import reload
        import Trimming.SigmoidGuessX0 as target_module
        import molass_legacy.KekLib.DebugPlot as plt

        def test_guess_x0_impl():
            reload(target_module)
            from molass_legacy.Trimming.SigmoidGuessX0 import guess_x0_impl
            guess_x0_impl(x, y)

        with plt.Dp(extra_button_specs=[("test_guess_x0_impl", test_guess_x0_impl)]):
            fig, ax = plt.subplots()
            ax.set_title("guess_x0 debug")
            ax.plot(x, y)
            fig.tight_layout()
            plt.show()

        reload(target_module)

    from molass_legacy.Trimming.SigmoidGuessX0 import guess_x0_impl
    i = guess_x0_impl(x, y, debug)
    return i

def guess_bent_sigmoid(x, y, x0=None, return_certainty=False, debug=False, save_fig=False):
    if x0 is None:
        x0 = guess_x0(x, y, debug=debug)

    nx = x - x0
    params = np.zeros((2,2))
    for k, part in enumerate([nx < 0, nx >= 0]):
        x_ = nx[part]
        y_ = y[part]
        slope, intercept = linregress(x_, y_)[0:2]
        if np.isnan(slope):
            # empty x_, y_
            params[k,:] = 0, y[0]
        else:
            params[k,:] = slope, intercept

    s1, s2 = params[:,0]
    b = params[0,1]
    L = params[1,1] - params[0,1]
    k = 1

    if return_certainty or debug or save_fig:
        sigm_y = bent_sigmoid(x, L, x0, k, b, s1, s2)
        rmsd = np.sqrt(np.mean((sigm_y - y)**2))
        height = np.max(y) - np.min(y)
        fc_certainty = np.exp(-rmsd/height*FC_CERTAINTY_SCALE)

    if debug or save_fig:
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy._MOLASS.SerialSettings import get_setting
        from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
        from time import sleep

        std = np.std(y)

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_bent_sigmoid debug for %s" % get_in_folder())
            ax.plot(x, y)
            ax.plot(x, sigm_y)
            tx = np.mean(ax.get_xlim())
            ymin, ymax = ax.get_ylim()
            ty = np.mean([ymin, ymax])
            dy = (ymax - ymin)*0.1
            ax.text(tx, ty+dy, "RMSD=%.3g" % (rmsd), ha="center", va="center", fontsize=20, alpha=0.5)
            ax.text(tx, ty, "RMSD/STD=%.3g" % (rmsd/std), ha="center", va="center", fontsize=20, alpha=0.5)
            ax.text(tx, ty-dy, "RMSD/HEIGHT=%.3g" % (rmsd/height), ha="center", va="center", fontsize=20, alpha=0.5)
            ax.text(tx, ty-2*dy, "CERTAINTY=%.3g" % (fc_certainty), ha="center", va="center", fontsize=20, alpha=0.5)
            fig.tight_layout()
            if debug:
                plt.show()
            else:
                assert save_fig
                # see https://stackoverflow.com/questions/14694408/runtimeerror-main-thread-is-not-in-main-loop
                # see al so DebugPlot.show()
                plt.switch_backend('agg')
                plt.show(block=False)
                debug_path = get_setting("debug_path")
                fig.savefig(debug_path)
                sleep(1)

    if return_certainty:
        return (L, x0, k, b, s1, s2), fc_certainty
    else:
        return L, x0, k, b, s1, s2

def fit_bent_sigmoid(x_, y_, x0):
    params0 = guess_bent_sigmoid(x_, y_, x0)
    popt, pcov = curve_fit(bent_sigmoid, x_, y_, params0)
    return popt, pcov

def ex_sigmoid(x, L ,x0, k, b, s1, s2):
    if s1 == 0 and s2 == 0:
        return sigmoid(x, L ,x0, k, b)
    else:
        return bent_sigmoid(x, L ,x0, k, b, s1, s2)

def ex_sigmoid_inv(y, L ,x0, k, b, s1, s2):
    if s1 == 0 and s2 == 0:
        return sigmoid_inv(y, L ,x0, k, b)
    else:
        return bent_sigmoid_inv(y, L ,x0, k, b, s1, s2)

def adjust_ex_sigmoid(x, y, params, debug=False):
    # introduced to cope with anomaly in 20200123_1 UV baseline

    L ,x0, k, b, s1, s2 = params[0:6]

    # optimize only b, s1, s2
    def objective(p, title=None):
        sy = ex_sigmoid(x, L ,x0, k, *p)
        if title is not None:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                ax.plot(x, sy)
                fig.tight_layout()
                plt.show()
        return np.sum((sy - y)**2)

    init_params = (b, s1, s2)
    if debug:
        objective(init_params, "adjust_ex_sigmoid: before minimize")
    ret = minimize(objective, init_params)
    if debug:
        objective(ret.x, "adjust_ex_sigmoid: after minimize")
    ret_params = params.copy()
    ret_params[3:6] = ret.x
    return ret_params



"""
    Selective.PeakProxy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

class PeakProxy:
    """
    This is a proxy class for EmgPeak class.
    the instance should behave consistently in FitRecord.sort_fit_recs,
    so that the order of the fit_recs is the same as the order of the peaks.
    task: to unify the code of PeakProxy and EmgPeakProxy.
    """
    def __init__(self, top_x=None, top_y=None, area_prop=None, sign=1):
        self.top_x = top_x
        self.top_y = top_y
        self.area_prop = area_prop
        self.sign = sign


"""
    Models.Stochastic.OligoporeImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from SecTheory.SecCF import sec_oligopore_phi_impl, gec_oligopore_phi_impl
from SecTheory.SecPDF import FftInvPdf

USE_DISPERSIVE = False

sec_oligopore_pdf = FftInvPdf(sec_oligopore_phi_impl)
gec_oligopore_pdf = FftInvPdf(gec_oligopore_phi_impl)

def oligopore_pdf(x, nt_pairs, pszp, x0, N0=None):
    if N0 is None:
        return gec_oligopore_pdf(x - x0, nt_pairs, pszp, 0)     # using x0 directly may induce divergence
    else:
        return sec_oligopore_pdf(x - x0, nt_pairs, pszp, 0, N0) # using x0 directly may induce divergence


"""
    Selective.NumComponents.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct
from Selective.V1ParamsAdapter import make_decomp_result_impl

def change_num_components_impl(advanced_frame, target_props=None, debug=False):
    if debug:    
        import CFSD.SimpleDecompose
        reload(CFSD.SimpleDecompose)
    from CFSD.SimpleDecompose import decompose_elution_simply
    if debug:
        print("change_num_components_impl")
    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    if debug:
        print("edm_text_from_editor_frame: ", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    if debug:
        with plt.Dp(window_title='Change Num Components', button_spec=["Close"]):
            fig, ax = plt.subplots()
            ax.set_title("before change", fontsize=16)
            ax.plot(x, y)
            for params in params_array:
                cy = model(fx, params)
                ax.plot(x, cy, ":")
            fig.tight_layout()
            plt.show()

    num_peaks = advanced_frame.num_components.get()

    if target_props is None:
        import Selective.BridgeUtils
        reload(Selective.BridgeUtils)
        from Selective.BridgeUtils import decompose_by_bridge
        traditional_info = Struct(num_peaks=num_peaks, mapper=editor_frame.mapper)
        decomp_result = decompose_by_bridge(fx, y, uv_y, model, traditional_info, debug=debug)
        if decomp_result is None:
            return
    else:
        if num_peaks < 3:
            traditional_info = Struct(num_peaks=num_peaks, mapper=editor_frame.mapper)
            decomp_result = decompose_elution_simply(fx, y, uv_y, model, traditional_info, props=target_props, debug=False)
        else:
            import Elution.Proportional
            reload(Elution.Proportional)
            from molass_legacy.Elution.Proportional import decompose_by_proportions
            params_result = decompose_by_proportions(fx, y, uv_y, model, target_props, xr_only=True, debug=debug)
            if params_result is None:
                return
            decomp_result = make_decomp_result_impl(editor, params_result, debug=False)

    with plt.Dp(window_title='Change Num Components', button_spec=["Accept", "Cancel"]):
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("after change", fontsize=20)
        ax1.set_title("UV", fontsize=16)
        ax1.plot(x, uv_y)
        cy_list = []
        for rec in decomp_result.opt_recs_uv:
            cy = rec.evaluator(fx)
            cy_list.append(cy)
            ax1.plot(x, cy, ":")
        ty = np.sum(cy_list, axis=0)
        ax1.plot(x, ty, ":", color="red")
        ax2.set_title("XR", fontsize=16)
        ax2.plot(x, y)
        cy_list = []
        for rec in decomp_result.opt_recs:
            cy = rec.evaluator(fx)
            cy_list.append(cy)
            ax2.plot(x, cy, ":")
        ty = np.sum(cy_list, axis=0)
        ax2.plot(x, ty, ":", color="red")
        fig.tight_layout()
        ret = plt.show()

    print("ret=", ret)
    if ret:
        advanced_frame.update_button_status(change_id="CNC")
        editor.update_current_frame_with_result(decomp_result, debug=True)


"""

    Models.ModelUtils.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF

"""
import re
import numpy as np
from molass_legacy.Decomposer.FitRecord import MIN_SEPARATE_AREA_RATIO, IGNORABLE_AREA_RATIO

model_name_re   = re.compile( r'Model\((\w+)\)' )

def _get_model_name( name ):
    m = model_name_re.match( name )
    if m:
        model_name  = m.group(1).upper()
    else:
        model_name  = None
    return model_name

plot_continue = True

def model_debug_plot(model, params, x_, y_, x, y, tau_hint, where_text, residual, before_params=None, print_callstack=False, work_y=None):
    """
    remember to make friendly this kind of debugging tools
    """
    import molass_legacy.KekLib.DebugPlot as plt
    global plot_continue
    if not plot_continue:
        return

    if print_callstack:
        from molass_legacy.KekLib.CallStack import CallStack
        print('model_debug_plot', CallStack())

    from molass_legacy.KekLib.OurMatplotlib import get_color
    plt.push()

    if before_params is None:
        params_list = [params]
    else:
        params_list = [before_params, params]

    ncols = len(params_list)
    figsize = (8*ncols, 7)
    fig, axes_ = plt.subplots(ncols=ncols, figsize=figsize)

    if ncols == 1:
        axes = [axes_]
        where_texts = [where_text]
    else:
        axes = axes_
        where_texts = ["before fit", "after fit"]

    for ax, params_, where_text_ in zip(axes, params_list, where_texts):
        model_name = model.get_name()
        model_text = "model: %s; " % model_name
        residual_text = " fitting for the residual" if residual else ""
        ax.set_title( model_text + where_text_ + '; tau_hint=' + str(tau_hint) + residual_text )
        ax.plot( x, y, ':', color='orange' )
        ax.plot( x_, y_, color='orange', label='data' )
        if work_y is not None:
            ax.plot(x, work_y, ':', label='work', color='green')
        color=get_color(0)
        ax.plot( x, model.eval( params_, x=x ), ':', color=color )
        ax.plot( x_, model.eval( params_, x=x_ ), color=color, label='model' )

        if model_name in ["EGHA", "EMGA"]:
            mu = params_[1]             # for EMG, EGH
        else:
            assert False
        tau = params_[3]
        ymin, ymax = ax.get_ylim()
        tx = mu
        ty = ymin*0.5 + ymax*0.5
        ax.text(tx, ty, "tau=%.3g" % tau, ha='center', fontsize=30, alpha=0.3)
        ax.legend()

    fig.tight_layout()
    ret = plt.show()
    if not ret:
        plot_continue = False
    plt.pop()

def plot_component_curves(ax, x, y, params_list, baseline, affine=False, color='blue', add_moments=False):
    from molass_legacy.Models.EGH import egh, egha    # placed here to avoid circular import
    model_func = egha if affine else egh
    ax.plot(x, y, color=color, label='data')
    cy_list = []
    for k, params in enumerate(params_list):
        cy = model_func(x, *params)
        cy_list.append(cy)
        ax.plot(x, cy, ':', label='component-%d' % k)
        if add_moments:
            from molass_legacy.Peaks.ElutionModels import compute_moments_from_egh_params
            M = compute_moments_from_egh_params(*params[1:4])
            ax.axvline(x=M[0], color="green")
    ty = np.sum(cy_list, axis=0)
    ax.plot(x, ty, ':', color="red", label='model total')
    ax.plot(x, baseline, color="red")
    ax.legend()

class PairedRangeProxy:
    def __init__(self, range_list):
        self.range_list = range_list
    
    def get_fromto_list(self):
        return self.range_list

def get_paired_ranges_from_params_array(model, x, params_array,
                                        separate_ratio=None,
                                        ignoreable_ratio=IGNORABLE_AREA_RATIO,
                                        return_indeces=False,
                                        want_num_components=None,
                                        select=None,
                                        debug=False):
    if separate_ratio is None:
        separate_ratio = MIN_SEPARATE_AREA_RATIO
    props = model.get_proportions(x, params_array)
    paired_ranges = []
    indeces = []
    if ignoreable_ratio is None or want_num_components is None:
        ignoreable_ratio_ = ignoreable_ratio
    else:
        num_peaks = len(props)
        sorted_props = sorted(props)
        n = num_peaks - want_num_components
        # ignoreable_ratio_ = min(ignoreable_ratio, sorted_props[n]*0.9)
        ignoreable_ratio_ = np.average(sorted_props[n-1:n+1]) if n > 0 else 0.0
        print("sorted_props=", sorted_props)
        print("ignoreable_ratio %g nas been replaced by %g to satisfy want_num_components=%d"
              % (ignoreable_ratio, ignoreable_ratio_, want_num_components) )

    if False:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("temp plot")
            ret = plt.show()
            if not ret:
                return

    for k, (params, prop) in enumerate(zip(params_array, props)):
        if select is None:
            if ignoreable_ratio_ is not None and prop < ignoreable_ratio_:
                continue
        else:
            if k not in select:
                continue

        indeces.append(k)
        y = model(x, params)
        m = np.argmax(y)
        max_y = y[m]
        try:
            f, t = [int(round(v)) for v in model.get_range(x, y, max_y=max_y)]
            if prop > separate_ratio:
                top_x = int(round(x[m]))
                range_list = [(f,top_x), (top_x,t)]
            else:
                range_list = [(f,t)]
            paired_ranges.append(PairedRangeProxy(range_list))
        except:
            import logging
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            logger = logging.getLogger(__name__)
            log_exception(logger, "get_paired_ranges_from_params_array")
            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                def get_range_except():
                    from importlib import reload
                    import Models.ModelUtilsException
                    reload(Models.ModelUtilsException)
                    from molass_legacy.Models.ModelUtilsException import get_range_exceptionally
                    return get_range_exceptionally(x, model, params)
                with plt.Dp(extra_button_specs=[("get_range_except", get_range_except)]):
                    fig, ax = plt.subplots()
                    ax.set_title("exception plot")
                    ax.plot(x, y)
                    plt.show()
 
            from molass_legacy.Models.ModelUtilsException import get_range_exceptionally
            f, t = get_range_exceptionally(x, model, params)
            paired_ranges.append(PairedRangeProxy([(f,t)]))
            logger.info("%-th paired range %s has been exceptionally computed", k, (f,t))

    if return_indeces:
        return paired_ranges, indeces
    else:
        return paired_ranges

def compute_cy_list(model, x, params_array):
    cy_list = []
    for params in params_array:
        cy = model(x, params)
        cy_list.append(cy)
    return cy_list

def compute_raw_moment1(x, y):
    return np.sum(x*y)/np.sum(y)    # raw moment

def compute_raw_moments1(x, y_list):
    return np.asarray([np.sum(x*y)/np.sum(y) for y in y_list])

def compute_area_props(cy_list):
    area_props = []
    for cy in cy_list:
        area_props.append(np.sum(cy))
    return np.array(area_props)/np.sum(area_props)

def data_dump(x, y, peaks, prefix=""):
    import os
    from molass_legacy._MOLASS.SerialSettings import get_setting
    temp_folder = get_setting("temp_folder")
    if temp_folder is None:
        temp_folder = os.getcwd()
    print("data_dump: temp_folder=", temp_folder)
    np.savetxt(os.path.join(temp_folder, prefix+"x.dat"), x)
    np.savetxt(os.path.join(temp_folder, prefix+"y.dat"), y)
    np.savetxt(os.path.join(temp_folder, prefix+"peaks.dat"), peaks)


# coding: utf-8
"""

    TrimmingDebugDnd.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import re
import tkinterDnD
from TkDebugDnd import TkDebugDnd
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from .TrimmingInfo import TrimmingInfo

class TrimmingDebugDnd(TkDebugDnd):
    def __init__(self, dialog):
        self.dialog = dialog
        self.ret_info_re = re.compile(r"ret_info=(\[.+\])")
        TkDebugDnd.__init__(self, dialog, nrows=2)

    def on_drop(self, event, i, j):
        print("on_drop: ", (i,j), event.data)

        yn = MessageBox.askyesno("Question",
                        "Redrawing %s with %s\n"
                        "Ok?" % (str((i,j)), event.data),
                        parent=self)
        if yn:
            print("redraw")
            ret_info = self.decode_ret_info(event.data)
            self.dialog.redraw(ret_info)

    def decode_ret_info(self, str_data):
        m = self.ret_info_re.search(str_data)
        ret_info = eval(m.group(1))
        return ret_info

    def on_drag(self, event, i, j):
        print("on_drag: ")
        ret_info = self.dialog.get_trimming_info()
        return (tkinterDnD.COPY, "DND_Text", "on_drag text: " + str((i,j)) + "\nret_info=" + str(ret_info))



"""
    circular_import_check.py
"""
import os
import sys
dir_ = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
print(dir_)
sys.path.append(dir_)
from molass_legacy.Models.RateTheory.EDM import guess_multiple, edm_impl


"""
    Optimizer.ParamSetType.FdEmgParamsSheet.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
from molass_legacy.Peaks.ElutionModels import emg
from .EghParamsSheet import EghParamsSheet

class FdEmgParamsSheet(EghParamsSheet):
    def __init__(self, parent, params, dsets, optimizer):
        EghParamsSheet.__init__(self, parent, params, dsets, optimizer, model_func=emg)

    def get_wanted_params(self, params, optimizer):
        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, seccol_params = optimizer.split_params_simple(params)
        dark_cell_j_list = [2]
        num_extra_addresses = xr_params.shape[0]
        Npc, tI = seccol_params[[0,2]]
        return xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, seccol_params, dark_cell_j_list, num_extra_addresses, (tI, Npc)



"""
Python Tkinter Splash Screen

This script holds the class SplashScreen, which is simply a window without
the top bar/borders of a normal window.

The window width/height can be a factor based on the total screen dimensions
or it can be actual dimensions in pixels. (Just edit the useFactor property)

Very simple to set up, just create an instance of SplashScreen, and use it as
the parent to other widgets inside it.

www.sunjay-varma.com
"""

"""
The original code has been taken from
    http://code.activestate.com/recipes/577271-tkinter-splash-screen/
and modified by Masatsuyo Takahashi, KEK-PF
"""
import tkinter      as Tk
import tkinter.ttk  as ttk

class SplashScreen(Tk.Frame):
    def __init__(self, master=None, width=0.8, height=0.6, useFactor=True):
        Tk.Frame.__init__(self, master)
        self.pack(side=Tk.TOP, fill=Tk.BOTH, expand=Tk.YES)

        # get screen width and height
        ws = self.master.winfo_screenwidth()
        hs = self.master.winfo_screenheight()
        w = (useFactor and ws*width) or width
        h = (useFactor and ws*height) or height
        # calculate position x, y
        x = (ws/2) - (w/2) 
        y = (hs/2) - (h/2)
        self.master.geometry('%dx%d+%d+%d' % (w, h, x, y))
        self.master.overrideredirect(True)
        self.lift()

def demo(root):
    sp = SplashScreen(root)
    sp.config(bg="#3366ff")

    m = Tk.Label(sp, text="This is a test of the splash screen\n\n\nThis is only a test.\nwww.sunjay-varma.com")
    m.pack(side=Tk.TOP, expand=Tk.YES)
    m.config(bg="#3366ff", justify=Tk.CENTER, font=("calibri", 29))
    
    Tk.Button(sp, text="Press this button to kill the program", bg='red', command=root.destroy).pack(side=Tk.BOTTOM, fill=Tk.X)
    root.mainloop()

if __name__ == '__main__':
    root = Tk.Tk()
    demo(root)



"""
    Simulative.DataBridgeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def get_databridge(decomp_editor, devel=True):
    from importlib import reload
    import Batch.DataBridge
    reload(Batch.DataBridge)
    from molass_legacy.Batch.DataBridge import DataBridge
    sd = decomp_editor.judge_holder.sd_orig
    pre_recog = decomp_editor.judge_holder.pre_recog
    params_array = decomp_editor.get_current_params_array()
    num_components = len(params_array)
    print("num_components =", num_components)
    bridge = DataBridge(sd, pre_recog)
    bridge.prepare_for_lrf_source(num_components=num_components)
    return bridge

def get_lrf_source_impl(decomp_editor):
    bridge = get_databridge(decomp_editor)
    return bridge.get_lrf_source()


"""
    Optimizer.Devel.TwinPeakMaker.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np

def make_twin_callback_txt(params_type, i, in_file, out_file):
    from molass_legacy.Optimizer.StateSequence import read_callback_txt_impl, save_opt_params
    fv_list, x_list = read_callback_txt_impl(in_file)
    fh = open(out_file, 'w')
    for j, (fv_rec, params) in enumerate(zip(fv_list, x_list)):
        # print([j], fv_rec, params)
        counter, fv, accept, time = fv_rec
        xr_params, xr_baseparams, rg_params, (a, b), uv_params, uv_baseparams, (c, d), sdmcol_params = params_type.split_params_simple(params)
        xr_v = xr_params[i]/2
        new_xr_params = np.concatenate([xr_params[:i], np.ones(2)*xr_v, xr_params[i+1:]])

        new_rg_params = np.concatenate([rg_params[:i], np.ones(2)*rg_params[i], rg_params[i+1:]])

        uv_v = uv_params[i]/2
        new_uv_params = np.concatenate([uv_params[:i], np.ones(2)*uv_v, uv_params[i+1:]])
        new_params = np.concatenate([new_xr_params, xr_baseparams, new_rg_params, (a,b), new_uv_params, uv_baseparams, (c, d), sdmcol_params])

        save_opt_params(fh, new_params, fv, accept, counter)
    fh.close()

if __name__ == '__main__':
    import sys
    this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.append(os.path.dirname(os.path.dirname(this_dir)))
    from ModelParams.SdmParams import SdmParams
    in_file = r"D:\TODO\20240917\C1015F-5\reports\analysis-004\optimized\jobs\063\callback.txt"
    out_file = r"D:\TODO\20240917\C1015F-5\reports\analysis-004\optimized\jobs\064\callback.txt"
    params_type = SdmParams(6)
    make_twin_callback_txt(params_type, 1, in_file, out_file)


"""
    from
    https://start-python.hateblo.jp/entry/2020/01/18/090000
"""
import random
import numpy as np
import matplotlib.pylab as plt
import matplotlib.animation as animation
from mpl_toolkits.mplot3d import Axes3D

# Qテーブルの初期化
q_table_1 = np.zeros((6, 6))
q_table_2 = np.zeros((6, 6))

# ３Dグラフの準備
fig = plt.figure()

x = [1, 2, 3, 4, 5, 6]
y = [1, 2, 3, 4, 5, 6]
X, Y = np.meshgrid(x, y)

a1 = random.randint(0, 5)
a0 = random.randint(0, 5)

def update(i):
    global a0, a1, a2

    plt.cla()

    a2 = a1
    a1 = a0
    a0 = random.randint(0, 5)

    if a0 == 1 or a0 == 3 or a0 == 5:
        q_table_1[a1, a2] += 1
    if a0 == 0 or a0 == 2 or a0 == 4:
        q_table_2[a1, a2] += 1

    Z1 = q_table_1
    Z2 = q_table_2

    ax = Axes3D(fig)
    
    plt.title('i=' + str(i) + '   [ ' + str(a0+1) + ' ]')

    ax.plot_wireframe(X, Y, Z1, color='b')
    ax.plot_wireframe(X, Y, Z2, color='r')

ani = animation.FuncAnimation(fig, update, interval = 100)
plt.show()



"""
    EghAdjuster.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt

class EghAdjuster:
    def __init__(self):
        pass

    def fit(self, fullopt, fv_array, x_list, debug=True):
        pass

    def convert(self, params):
        return params



"""
    SigmoidChain.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from .Sigmoid import fit_bent_sigmoid, ex_sigmoid

HW = 50
JOINT_PENALTY_SCALE = 1e5

class SigmoidChain:
    def __init__(self, x, y, ppk, debug=False):
        self.x = x
        self.y = y
        self.ppk = ppk
        init_params = []
        split_slices = []

        sorted_ppk = sorted(ppk)
        if debug:
            print("ppk=", ppk, "sorted_ppk=", sorted_ppk)

        start = 0
        last_p = None
        for p in sorted_ppk:
            slice_ = slice(max(0, p - HW), min(len(x), p + HW))
            x_ = x[slice_]
            y_ = y[slice_]
            try:
                popt, pcov = fit_bent_sigmoid(x_, y_, p)
                init_params.append(popt)
                p_ = popt[1]
                if last_p is not None:
                    stop = int(round((last_p + p_)/2))
                    split_slices.append(slice(start, stop))
                    start = stop
                last_p = p_
            except:
                log_exception(None, "fit_bent_sigmoid failure: ")

        split_slices.append(slice(start, None))
        self.split_slices = split_slices

        major_params = []
        slope_params = []
        last_s2 = None
        for params in init_params:
            major_params.append(params[0:4])
            s1, s2 = params[4:]
            if last_s2 is None:
                slope_params.append(s1)
            else:
                slope_params.append((last_s2+s1)/2)
            last_s2 = s2
        slope_params.append(last_s2)

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("before minimize")
                ax.plot(x, y)
                ax.plot(x, self.evaluate(x, slope_params))
                fig.tight_layout()
                plt.show()

        def obj_func(p):
            y_ = np.zeros(len(x))
            joint_penalty = 0
            for k, slice_ in enumerate(self.split_slices):
                m_param = major_params[k]
                s_param = p[k:k+2]
                y_[slice_] += ex_sigmoid(x[slice_], *m_param, *s_param)
                if k > 0:
                    i = slice_.start
                    joint_penalty += JOINT_PENALTY_SCALE*(y_[i-1] - y_[i])**2
            return np.sum((y_ - y)**2) + joint_penalty

        result = minimize(obj_func, slope_params)
        popt = result.x

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("after minimize")
                ax.plot(x, y)
                ax.plot(x, self.evaluate(x, popt))
                fig.tight_layout()
                plt.show()

        self.major_params = major_params
        self.slope_params = popt

    def evaluate(self, x, slope_params=None):
        if slope_params is None:
            slope_params = self.slope_params

        y = np.zeros(len(x))
        k = 0
        for slice_, m_param in zip(self.split_slices, self.major_params):
            y[slice_] += ex_sigmoid(x[slice_], *m_param, *slope_params[k:k+2])
            k += 1
        return y

    def __call__(self, x):
        return self.evaluate(x, self.slope_params)

    def plot_segments(self, ax, label=None, **kwargs):
        k = 0
        for slice_, m_param in zip(self.split_slices, self.major_params):
            x_ = self.x[slice_]
            label_ = label if label is None else label % k
            y_ = ex_sigmoid(x_, *m_param, *self.slope_params[k:k+2])
            ax.plot(x_, y_, label=label_, **kwargs)
            k += 1



# coding: utf-8
"""
    Rgg.KernelDensityDemo.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from pomegranate import GaussianKernelDensity
from .KernelDensityFix import UniformKernelDensity
import molass_legacy.KekLib.DebugPlot as plt

def demo():
    x = np.linspace(0, 50, 1000)

    fig, axes = plt.subplots(ncols=2, figsize=(14,7))

    for k, args in enumerate([
        (GaussianKernelDensity([30], bandwidth=1), 'Gaussian Kernel Density'),
        (UniformKernelDensity([30], bandwidth=1), 'Uniform Kernel Density'),
        ]):
        ax = axes[k]
        d, name = args
        ax.set_title(name)
        y = d.probability(x)
        ax.fill_between(x, 0, y, color='c')

    fig.tight_layout()
    plt.show()



"""
    AveragingDiscussion.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import copy
import numpy as np

def extrapolate(M, c, rank=2):
    U, s, VT = np.linalg.svd(M)
    M_  = U[:,0:rank] @ np.diag(s[0:rank]) @ VT[0:rank,:]
    c_  = c/np.max(c)
    C   = np.array([c_, c_**2])
    """
        P・C = M
        P = M・Cpinv
    """
    Cpinv = np.linalg.pinv(C)
    P   = M_ @ Cpinv
    return P

class AveragingDiscussion:
    def __init__(self, sd):
        import molass_legacy.KekLib.DebugPlot as plt
        self.mouse_axis = None

        self.q = sd.intensity_array[0,:,0]
        self.data = sd.intensity_array[:,:,1].T
        self.e_curve = sd.xray_curve

        peak_i = self.e_curve.primary_peak_i
        xslice = slice(0,100)
        yslice = slice(peak_i-50, peak_i+50)
        wslice = slice(peak_i-20, peak_i)

        x = self.q[xslice]
        y = np.arange(self.data.shape[1])[yslice]
        xx, yy = np.meshgrid( x, y )

        i = np.arange(xslice.start, xslice.stop)
        j = np.arange(yslice.start, yslice.stop)
        ii, jj = np.meshgrid( i, j )

        zz = self.data[ii, jj]

        self.fig = fig = plt.figure(figsize=(30, 8))
        fig.canvas.mpl_connect( 'draw_event', self.on_draw )
        fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )
        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132, projection='3d')
        ax3 = fig.add_subplot(133, projection='3d')
        self.axes = axes = [ax1, ax2, ax3]

        ax1.set_title("Five-line Average along Elutional Axis", fontsize=20)
        ax2.set_title("Extrapolation from 20 Five-line Averages", fontsize=20)
        ax3.set_title("Extrapolation from 20 raw lines", fontsize=20)

        for ax in axes:
            ax.plot_surface(xx, yy, zz, color='green', alpha=0.1 )
            x_ = np.ones(len(y))*0.02
            z_ = self.e_curve.y[yslice]
            ax.plot(x_, y, z_, color='orange', linewidth=3)

        for k in range(70, 75):
            y_ = np.ones(len(x))*k
            z_ = self.data[xslice,k]
            ax1.plot(x, y_, z_, color='pink')

        k = 72
        y_ = np.ones(len(x))*k
        z_ = np.average(self.data[xslice,70:75], axis=1)
        ax1.plot(x, y_, z_, color='yellow', linewidth=3, label='5L average')
        ax1.legend(bbox_to_anchor=(1, 0), loc='lower right', fontsize=20)

        def comparison_plot(ax, data, conc_y, color, alpha):
            M   = data[xslice, wslice]
            c   = conc_y[wslice]
            P   = extrapolate(M, c)
            w   = P[:,0]

            for k in range(wslice.start, wslice.stop):
                y_ = np.ones(len(x))*k
                z_ = data[xslice,k]
                ax.plot(x, y_, z_, color=color, alpha=alpha)

            k = peak_i-10
            y_ = np.ones(len(x))*k
            z_ = np.average(data[xslice,peak_i-20:peak_i], axis=1)
            ax.plot(x, y_, z_, color='cyan', linewidth=3, label='20L average')

            y_ = np.ones(len(x))*peak_i
            ax.plot(x, y_, w, color='red', linewidth=3, label='extrapolated')
            zs = z_ * w[-1]/z_[-1]
            ax.plot(x, y_, zs, color='lime', linewidth=3, label='20L average scaled')

            ax.legend(bbox_to_anchor=(1, 0), loc='lower right', fontsize=20)

        averaged_data = copy.deepcopy(self.data)
        averaged_c = copy.deepcopy(self.e_curve.y)
        n = 5
        m = n//2
        for j in range(self.data.shape[1]-n):
            averaged_data[:,j+m] = np.average(self.data[:,j:j+n], axis=1)
            averaged_c[j+m] = np.average(self.e_curve.y[j:j+n])

        comparison_plot(ax2, averaged_data, averaged_c,'yellow', 1)
        comparison_plot(ax3, self.data, self.e_curve.y, 'green', 0.5)

        fig.tight_layout()
        plt.show()

    def on_button_press(self, event):
        if event.inaxes in self.axes:
            self.mouse_axis = event.inaxes

    def on_draw(self, event):
        if self.mouse_axis is None:
            return

        import molass_legacy.KekLib.DebugPlot as plt
        ax1, ax2, ax3 = self.axes
        if self.mouse_axis == ax1:
            this_ax, oppo_axes = ax1, [ax2, ax3]
        elif self.mouse_axis == ax2:
            this_ax, oppo_axes = ax2, [ax1, ax3]
        elif self.mouse_axis == ax3:
            this_ax, oppo_axes = ax3, [ax1, ax2]
        else:
            return

        # TODO: investigate the cause of delayed synchronization
        self.mouse_axis = None
        for ax in oppo_axes:
            ax.view_init( this_ax.elev, this_ax.azim )
        plt.dp.mpl_canvas.draw()



# coding: utf-8
"""

    ファイル名：   CommandLineOptions.py

    処理内容：

        コマンドライン・オプションの取得

"""
import sys

from optparse import OptionParser

usage = "usage: autorg_kek [options] <datafile(s)>"

parser = OptionParser( usage=usage )

parser.add_option('', '--mininterval', action='store', default=3,
                help="Minimum acceptable Guinier interval length in points. (default: 3)")

parser.add_option('', '--smaxrg', action='store', default=1.3,
                help="Maximum acceptable SmaxRg value. (default: 1.3)")

parser.add_option('', '--sminrg', action='store', 
                help="Minimum acceptable SminRg value. (default unlimited)")

parser.add_option('-o', '--output', action='store',
                help="Relative or absolute path to save result" )

parser.add_option('-f', '--format', action='store',
                help="Output format, either of: csv, ssv")

parser.add_option('-A', '--ATSAS', action='store_true', 
                help="output ATSAS autorg columns only")

parser.add_option('-r', '--robust', action='store_true',
                help="output result even if it is quite bad")

parser.add_option('-S', '--server', action='store_true',
                help="run as a server")

parser.add_option('-v', '--version', action='store_true', 
                help="show the appication's version info")



"""

    ファイル名：   SpecialistOptions.py

    処理内容：

        専門家者用の設定変更ダイアログ

    Copyright (c) 2017-2023, SAXS Team, KEK-PF

"""

import os
import re
import warnings
import time

from molass_legacy.KekLib.BasicUtils             import get_filename_extension
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog, ttk, Font, FileDialog, is_empty_val, checkFolder
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
import OurMessageBox        as MessageBox
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting, ITEM_DEFAULTS
from DevSettings            import get_dev_setting, set_dev_setting, ITEM_DEFAULTS as DEV_ITEM_DEFAULTS
from molass_legacy.KekLib.TkCustomWidgets        import FolderEntry

ENABLE_Rayleigh_Scatter_OPTION  = False
ENABLE_XRAY_PROP_CONC_SMOOTHING = False
ENABLE_REGRESSION_LOOKING_BACK  = False
ENABLE_Mapping_Animation        = False
ENABLE_DEPRECATED_OPTIONS       = False

class SpecialistOptionsDialog( Dialog ):
    def __init__( self, parent, title, grand_parent=None ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.grand_parent       = grand_parent
        self.title_             = title
        self.applied            = False
        self.temp_absorbance    = None
        self.absorbance_has_been_changed    = False

    def show( self ):
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__(self, self.parent, self.title_ )

        self.parent.config( cursor='' )

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self )

        secondary_padx  = 20

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=10, pady=20 )

        grid_row = 0
        spacing_frame = Tk.Frame( iframe, width=80 )
        spacing_frame.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Abnormal Data Handling
        # ----------------------------------------------------------------------
        grid_row += 1
        abnormal_data_label = Tk.Label( iframe, text='Abnormal Data Handling' )
        abnormal_data_label.grid( row=grid_row, column=0, sticky=Tk.W )

        # enable mapping animation
        grid_row += 1
        self.data_exclusion = Tk.IntVar()
        self.data_exclusion.set( get_setting( 'data_exclusion' ) )

        cb = Tk.Checkbutton( iframe, text="enable data exclusion indication in the Data File Table",
                                variable=self.data_exclusion  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Concentration Mapping Option
        # ----------------------------------------------------------------------
        if ENABLE_Mapping_Animation:
            grid_row += 1
            mapping_options_label = Tk.Label( iframe, text='Concentration Mapping Option' )
            mapping_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

            # enable mapping animation
            grid_row += 1
            self.enable_mapping_anim = Tk.IntVar()
            self.enable_mapping_anim.set( get_setting( 'enable_mapping_anim' ) )

            cb = Tk.Checkbutton( iframe, text="enable Mapping Animation",
                                    variable=self.enable_mapping_anim  )
            cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

            grid_row += 1
            spacing = Tk.Label( iframe, text='' )
            spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Data Correction Options
        # ----------------------------------------------------------------------
        grid_row += 1
        correction_options_label = Tk.Label( iframe, text='Data Correction Options' )
        correction_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

        # enable mapping animation
        grid_row += 1
        self.enable_xb_save = Tk.IntVar()
        self.enable_xb_save.set( get_dev_setting( 'enable_xb_save' ) )

        cb = Tk.Checkbutton( iframe, text="enable X-ray Scattering Baseline Save",
                                variable=self.enable_xb_save  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

        self.xb_folder = Tk.StringVar()
        # analysis_folder = get_setting( 'analysis_folder' )
        _, analysis_folder = self.parent.make_analysis_folder()
        if analysis_folder is not None:
            folder_path = os.path.join( analysis_folder, 'xray_scattering_base' ).replace( '\\', '/' )
            self.xb_folder.set( folder_path )
        folder_entry = FolderEntry( iframe, textvariable=self.xb_folder, width=70,
                                            on_entry_cb=self.on_entry_xb_folder )
        folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        grid_row += 1
        if len(self.parent.file_info_table.datafiles) > 0:
            self.filename_example   = os.path.split( self.parent.file_info_table.datafiles[0] )[-1]
            self.filename_extention = '.' + get_filename_extension( self.filename_example )
        else:
            self.filename_example   = None
            self.filename_extention = ".dat"

        postfix_frame = Tk.Frame( iframe )
        postfix_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        postfix_label = Tk.Label( postfix_frame, text="filename postfix" )
        postfix_label.grid( row=0, column=0 )
        self.base_file_postfix = Tk.StringVar()
        self.base_file_postfix.set( "_base" )

        self.base_file_postfix_entry = Tk.Entry( postfix_frame, textvariable=self.base_file_postfix, width=10 )
        self.base_file_postfix_entry.grid( row=0, column=1, padx=5 )

        as_in_label = Tk.Label( postfix_frame, text="as in" )
        as_in_label.grid( row=0, column=2 )

        self.postfix_eg = Tk.StringVar()
        self.postfix_eg_update()
        eg_label = Tk.Label( postfix_frame, textvariable=self.postfix_eg )
        eg_label.grid( row=0, column=3, padx=3 )

        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Elution Mapping Options
        # ----------------------------------------------------------------------
        grid_row += 1
        mapping_options_label = Tk.Label( iframe, text='Elution Mapping Options' )
        mapping_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

        # Peak Mapping Only
        grid_row += 1
        self.peak_mapping_only = Tk.IntVar()
        self.peak_mapping_only.set( get_setting('peak_mapping_only') )

        cb = Tk.Checkbutton( iframe, text="do mapping with peak tops and end points only",
                                variable=self.peak_mapping_only  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

        # enable LPM variations (linear, quadratic, spline)
        grid_row += 1
        self.lpm_variations = Tk.IntVar()
        self.lpm_variations.set( get_setting('lpm_variations') )

        cb = Tk.Checkbutton( iframe, text="enable LPM variations: linear, quadratic, spline",
                                variable=self.lpm_variations  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        # allow angular slope in MF-baseplane
        grid_row += 1
        self.allow_angular_slope_in_mf = Tk.IntVar()
        self.allow_angular_slope_in_mf.set( get_setting('allow_angular_slope_in_mf') )

        cb = Tk.Checkbutton( iframe, text="allow angular slope in MF-baseplane  ( i.e., z = a*x + b*y + c, a != 0 )",
                                variable=self.allow_angular_slope_in_mf  )
        cb.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=secondary_padx )

        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Elution Curve Modeling and Elution Decomposer Options
        # ----------------------------------------------------------------------
        grid_row += 1
        guinier_options_label = Tk.Label( iframe, text='Elution Curve Modeling and Elution Decomposer Options' )
        guinier_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

        # enable Affine Transformation
        grid_row += 1
        self.enable_affine_tran = Tk.IntVar()
        self.enable_affine_tran.set( get_setting( 'enable_affine_tran' ) )

        cb = Tk.Checkbutton( iframe, text="enable Affine Transformation",
                                variable=self.enable_affine_tran  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

        # enable Decomposition from molass_legacy.UV-Xray separation
        grid_row += 1
        self.decomp_from_separation = Tk.IntVar()
        self.decomp_from_separation.set( get_setting( 'decomp_from_separation' ) )

        cb = Tk.Checkbutton( iframe, text="enable Decomposition from molass_legacy.UV-Xray separation",
                                variable=self.decomp_from_separation  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx   )

        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Guinier Analysis Options
        # ----------------------------------------------------------------------
        dataset_is_ready = self.parent.dataset_is_ready
        self.serial_data = self.parent.serial_data
        ranges = [ rec for rec in self.serial_data.xray_curve.peak_info ] if dataset_is_ready else []
        self.init_ranges = [ [ 1, tuple(range_) ]  for range_ in ranges ]
        self.num_peaks = len(ranges)

        grid_row += 1
        guinier_options_label = Tk.Label( iframe, text='Guinier Analysis Options' )
        guinier_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        self.fixed_guinier_start = Tk.IntVar()
        self.fixed_guinier_start.set( get_setting( 'fixed_guinier_start' ) )

        state = Tk.NORMAL if dataset_is_ready else Tk.DISABLED
        cb = Tk.Checkbutton( iframe, text="Fix Guinier interval start point to the",
                                variable=self.fixed_guinier_start, state=state )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        self.guinier_start_dependents = []

        frame_ = Tk.Frame( iframe )
        frame_.grid( row=grid_row, column=1, sticky=Tk.W )

        self.guinier_start_point = Tk.IntVar()
        self.guinier_start_point.set( 0 )
        entry_ = Tk.Spinbox( frame_, textvariable=self.guinier_start_point,
                                            from_=0, to=50, increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        entry_.pack( side=Tk.LEFT  )
        self.guinier_start_dependents.append( entry_ )

        label_ = Tk.Label( frame_, text=' -th point; looking at the scattering curve on the' )
        label_.pack( side=Tk.LEFT )
        self.guinier_start_dependents.append( label_ )

        self.guinier_fig_peak_no = Tk.IntVar()
        self.guinier_fig_peak_no.set( 0 )

        entry_ = Tk.Spinbox( frame_, textvariable=self.guinier_fig_peak_no,
                                            from_=0, to=max(1, self.num_peaks-1), increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        entry_.pack( side=Tk.LEFT  )
        self.guinier_start_dependents.append( entry_ )

        label_ = Tk.Label( frame_, text=' -th peak' )
        label_.pack( side=Tk.LEFT )
        self.guinier_start_dependents.append( label_ )

        self.guinier_btn = Tk.Button( iframe, text="Guinier plot", command=self.guinier_plot_dialog )
        self.guinier_btn.grid( row=grid_row, column=2, sticky=Tk.W )
        self.guinier_start_dependents.append( self.guinier_btn )

        self.fixed_guinier_start_tracer()   # set initial state
        self.fixed_guinier_start.trace( 'w', self.fixed_guinier_start_tracer )

        # ----------------------------------------------------------------------
        # Extrapolation Options
        # ----------------------------------------------------------------------
        grid_row += 1
        extrapolation_options_label = Tk.Label( iframe, text='Extrapolation Options' )
        extrapolation_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

        enable_new_features = get_setting('enable_new_features')

        # enable Concentration Datatype Changes
        grid_row += 1
        self.enable_conctype_change = Tk.IntVar()
        self.enable_conctype_change.set(get_setting('enable_conctype_change'))
        if enable_new_features:
            cb = Tk.Checkbutton( iframe, text="enable Concentration Datatype Changes",
                                    variable=self.enable_conctype_change  )
            cb.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=secondary_padx  )

        # Extended Concentration Dependency
        grid_row += 1
        self.extended_conc_dep = Tk.IntVar()
        self.extended_conc_dep.set( get_setting( 'extended_conc_dep' ) )
        cb = Tk.Checkbutton( iframe, text="enable Extended Concentration Dependency  ( i.e., I(q) = A(q)*c + B(q)*c² + Z(q)*c³ )",
                                variable=self.extended_conc_dep  )
        cb.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=secondary_padx )

        # Rank Variation
        grid_row += 1
        self.allow_rank_variation = Tk.IntVar()
        self.allow_rank_variation.set( get_setting( 'allow_rank_variation' ) )
        cb = Tk.Checkbutton( iframe, text="allow Rank Variation  ( i.e., no SVD denoise, additional rank allowance )",
                                variable=self.allow_rank_variation  )
        cb.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=secondary_padx  )

        # use xray-transformed concentration
        if False:
            grid_row += 1
            self.use_xray_conc = Tk.IntVar()
            self.use_xray_conc.set( get_setting( 'use_xray_conc' ) )
            self.use_xray_conc.trace( 'w', self.use_xray_conc_tracer )

            column_ = 0
            cb = Tk.Checkbutton( iframe, text="use Xray-proportional concentration",
                                    variable=self.use_xray_conc  )
            cb.grid( row=grid_row, column=column_, sticky=Tk.W, padx=secondary_padx  )

            if ENABLE_XRAY_PROP_CONC_SMOOTHING:
                column_ += 1
                self.smoothed_xray_conc = Tk.IntVar()
                self.smoothed_xray_conc.set( get_dev_setting( 'smoothed_xray_conc' ) )

                cb = Tk.Checkbutton( iframe, text="apply smoothing to the left-mentioned concentration",
                                        variable=self.smoothed_xray_conc  )
                cb.grid( row=grid_row, column=column_, sticky=Tk.W  )

            self.apply_backsub = Tk.IntVar()
            self.apply_backsub.set( get_setting( 'apply_backsub' ) )

            column_ += 1
            self.apply_backsub_cb = Tk.Checkbutton( iframe, text="apply backsub by LPM",
                                    variable=self.apply_backsub  )
            self.apply_backsub_cb.grid( row=grid_row, column=column_, sticky=Tk.W  )
            self.use_xray_conc_tracer()

        # Recompute Regression Boundary by looking back
        if ENABLE_REGRESSION_LOOKING_BACK:
            grid_row += 1
            self.recompute_regboundary = Tk.IntVar()
            self.recompute_regboundary.set( get_dev_setting( 'recompute_regboundary' ) )

            cb = Tk.Checkbutton( iframe, text="recompute Regression Boundary by looking back",
                                    variable=self.recompute_regboundary  )
            cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx  )

        # almerge analyzer
        grid_row += 1
        self.almerge_analyzer = Tk.IntVar()
        self.almerge_analyzer.set( get_setting( 'almerge_analyzer' ) )

        cb = Tk.Checkbutton( iframe, text="enable Almerge Analyzer",
                                variable=self.almerge_analyzer, state=Tk.DISABLED  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx  )

        # space
        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        if ENABLE_DEPRECATED_OPTIONS:
            # ----------------------------------------------------------------------
            # Deprecated Extrapolation Options
            # ----------------------------------------------------------------------
            self.fixed_font = Font.Font( family="Courier", size=9 )
            extrapolation_state = Tk.NORMAL

            grid_row += 1
            extrapolation_options_label = Tk.Label( iframe, text='Deprecated Extrapolation Options' )
            extrapolation_options_label.grid( row=grid_row, column=0, sticky=Tk.W )

            # Width of Regression in Q-axis
            grid_row += 1
            angle_boundary_label = Tk.Label( iframe, text='Width of Regression in the direction of Q-axis', state=extrapolation_state )
            angle_boundary_label.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx  )

            regression_method_frame = Tk.Frame( iframe )
            regression_method_frame.grid( row=grid_row, column=1, sticky=Tk.W )

            self.regression_method_buttons = []
            self.regression_method = Tk.IntVar()
            self.regression_method.set( get_setting( 'zx_num_q_points' ) )

            b = Tk.Radiobutton( regression_method_frame, text='Five Points',
                        font=self.fixed_font,
                        variable=self.regression_method, value=5,
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=0, sticky=Tk.E )
            self.regression_method_buttons.append( b )

            b = Tk.Radiobutton( regression_method_frame, text='Three Points',
                        font=self.fixed_font,
                        variable=self.regression_method, value=3,
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=1, sticky=Tk.E, padx=27 )
            self.regression_method_buttons.append( b )

            b = Tk.Radiobutton( regression_method_frame, text='One Point',
                        font=self.fixed_font,
                        variable=self.regression_method, value=1,
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=2, sticky=Tk.E )
            self.regression_method_buttons.append( b )

            # Regression Boundary in A(q)
            grid_row += 1
            zx_boundary_label = Tk.Label( iframe, text='Regression Boundary', state=extrapolation_state )
            zx_boundary_label.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx  )

            zx_boundary_frame = Tk.Frame( iframe )
            zx_boundary_frame.grid( row=grid_row, column=1, sticky=Tk.W )

            self.zx_boundary_method_buttons = []
            self.zx_boundary_method = Tk.StringVar()
            self.zx_boundary_method.set( get_setting( 'zx_boundary_method' ) )

            b = Tk.Radiobutton( zx_boundary_frame, text='Automatic   ',
                        font=self.fixed_font,
                        variable=self.zx_boundary_method, value='AUTO',
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=0, sticky=Tk.E )
            self.zx_boundary_method_buttons.append( b )

            b = Tk.Radiobutton( zx_boundary_frame, text='Fixed at',
                        font=self.fixed_font,
                        variable=self.zx_boundary_method, value='FIXED',
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=1, sticky=Tk.E, padx=20 )
            self.zx_boundary_method_buttons.append( b )

            self.zx_boundary = Tk.DoubleVar()
            self.zx_boundary.set( get_setting( 'zx_boundary' ) )
            self.zx_boundary_entry = Tk.Entry( zx_boundary_frame, textvariable=self.zx_boundary, justify=Tk.CENTER, width=6 )
            self.zx_boundary_entry.grid( row=0, column=2, sticky=Tk.W )

            b = Tk.Radiobutton( zx_boundary_frame, text='No Boundary',
                        font=self.fixed_font,
                        variable=self.zx_boundary_method, value='NO',
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=3, sticky=Tk.E, padx=30 )
            self.zx_boundary_method_buttons.append( b )
            self.zx_boundary_method.trace( 'w', self.zx_boundary_method_tracer )

            # Construction Method in Wide Angle Region of A(q)
            grid_row += 1
            wide_angle_region_label = Tk.Label( iframe, text='A(q) Curve Construction in B(q)-Extinguished Region', state=extrapolation_state )
            wide_angle_region_label.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

            wide_angle_region_frame = Tk.Frame( iframe )
            wide_angle_region_frame.grid( row=grid_row, column=1, sticky=Tk.W )

            self.zx_build_method_buttons = []
            self.zx_build_method = Tk.StringVar()
            self.zx_build_method.set( get_setting( 'zx_build_method' ) )

            b = Tk.Radiobutton( wide_angle_region_frame, text='Use max conc. only',
                        variable=self.zx_build_method, value='MAX',
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=0, sticky=Tk.E )
            self.zx_build_method_buttons.append( b )

            b = Tk.Radiobutton( wide_angle_region_frame, text='Conc-weighted regression',
                        variable=self.zx_build_method, value='REG',
                        state=extrapolation_state,
                        )
            b.grid( row=0, column=1, sticky=Tk.E, padx=2 )
            self.zx_build_method_buttons.append( b )

            # space
            grid_row += 1
            spacing = Tk.Label( iframe, text='' )
            spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Software Availability
        # ----------------------------------------------------------------------
        grid_row += 1
        label = Tk.Label( iframe, text='Software Availability' )
        label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        self.revoke_excel = Tk.IntVar()
        self.revoke_excel_init = get_setting('revoke_excel')
        self.revoke_excel.set(self.revoke_excel_init)

        cb = Tk.Checkbutton( iframe,
                                        text='revoke Excel Availability',
                                        variable=self.revoke_excel,
                                        state=Tk.NORMAL )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        grid_row += 1
        self.revoke_atsas = Tk.IntVar()
        self.revoke_atsas_init = get_setting('revoke_atsas')
        self.revoke_atsas.set(self.revoke_atsas_init)

        cb = Tk.Checkbutton( iframe,
                                        text='revoke ATSAS Availability',
                                        variable=self.revoke_atsas,
                                        state=Tk.NORMAL )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        grid_row += 1
        self.revoke_cuda = Tk.IntVar()
        self.revoke_cuda_init = get_setting('revoke_cuda')
        self.revoke_cuda.set(self.revoke_cuda_init)

        cb = Tk.Checkbutton( iframe,
                                        text='revoke CUDA Availability',
                                        variable=self.revoke_cuda,
                                        state=Tk.NORMAL )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        # space
        grid_row += 1
        spacing = Tk.Label( iframe, text='' )
        spacing.grid( row=grid_row, column=0 )

        # ----------------------------------------------------------------------
        # Devolopment/Maintenance
        # ----------------------------------------------------------------------
        grid_row += 1
        maintenance = Tk.Label( iframe, text='Maintenance' )
        maintenance.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        self.maintenance_mode = Tk.IntVar()
        self.maintenance_mode.set( get_setting( 'maintenance_mode' ) )

        maintenance_mode_cb = Tk.Checkbutton( iframe,
                                        text='enable Maintenance Mode',
                                        variable=self.maintenance_mode,
                                        state=Tk.NORMAL )
        maintenance_mode_cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=secondary_padx )

        # ----------------------------------------------------------------------
        # Reset to Defaults button
        # ----------------------------------------------------------------------
        grid_row += 1
        reset_button = Tk.Button( iframe, text="Reset to Defaults", command=self.reset_to_defaults )
        reset_button.grid( row=grid_row, column=2, sticky=Tk.E )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

        self.update()

    def buttonbox( self ):
        box = Tk.Frame( self )
        box.pack( pady=10 )
        Dialog.buttonbox( self, frame=box )

    def on_entry_xb_folder( self ):
        pass

    def postfix_eg_update( self ):
        try:
            filename = self.filename_example.replace( self.filename_extention, self.base_file_postfix.get() + self.filename_extention )
            self.postfix_eg.set( filename )
        except:
            pass

    def use_xray_conc_tracer( self, *args ):
        if self.use_xray_conc.get() == 0:
            state = Tk.DISABLED
        else:
            state = Tk.NORMAL
        self.apply_backsub_cb.config( state=state )

    def show_absorbance_figure( self ):
        if ENABLE_Rayleigh_Scatter_OPTION:
            set_setting( 'consider_scatter',    self.consider_scatter.get() )
        set_setting( 'scatter_picking',     self.zero_absorbance.get() )

        from molass_legacy.UV.AbsorbancePlot import show_absorbance_figure_util
        show_absorbance_figure_util( self, make_temp=True )

    def reset_to_defaults( self ):
        self.data_exclusion.set(        ITEM_DEFAULTS['data_exclusion'] )
        self.enable_affine_tran.set(    ITEM_DEFAULTS['enable_affine_tran'] )
        self.decomp_from_separation.set(ITEM_DEFAULTS['decomp_from_separation'] )
        if ENABLE_Rayleigh_Scatter_OPTION:
            self.consider_scatter.set(  ITEM_DEFAULTS['consider_scatter'] )
        self.almerge_analyzer.set(      ITEM_DEFAULTS['almerge_analyzer'] )
        self.peak_mapping_only.set(     ITEM_DEFAULTS['peak_mapping_only'] )
        self.lpm_variations.set(        ITEM_DEFAULTS['lpm_variations'] )
        self.allow_angular_slope_in_mf.set( ITEM_DEFAULTS['allow_angular_slope_in_mf'] )

        if ENABLE_DEPRECATED_OPTIONS:
            self.regression_method.set(         ITEM_DEFAULTS['zx_num_q_points' ] )
            self.zx_build_method.set(           ITEM_DEFAULTS['zx_build_method' ] )
            self.zx_boundary_method.set(        ITEM_DEFAULTS['zx_boundary_method' ] )
            self.zx_boundary.set(               ITEM_DEFAULTS['zx_boundary' ] )

        self.revoke_excel.set(          ITEM_DEFAULTS['revoke_excel'] )
        self.revoke_atsas.set(          ITEM_DEFAULTS['revoke_atsas'] )
        self.revoke_cuda.set(           ITEM_DEFAULTS['revoke_cuda'] )
        self.maintenance_mode.set(      ITEM_DEFAULTS['maintenance_mode'] )

        if ENABLE_Mapping_Animation:
            self.enable_mapping_anim.set(   ITEM_DEFAULTS['enable_mapping_anim' ] )

        self.allow_rank_variation.set(  ITEM_DEFAULTS['allow_rank_variation' ] )
        self.enable_conctype_change.set(    ITEM_DEFAULTS['enable_conctype_change' ] )
        self.extended_conc_dep.set(     ITEM_DEFAULTS['extended_conc_dep' ] )
        if False:
            self.use_xray_conc.set(         ITEM_DEFAULTS['use_xray_conc' ] )
            self.apply_backsub.set(         ITEM_DEFAULTS['apply_backsub'] )
            if ENABLE_XRAY_PROP_CONC_SMOOTHING:
                self.smoothed_xray_conc.set(    DEV_ITEM_DEFAULTS['smoothed_xray_conc' ] )
        # self.zx_add_constant.set(       DEV_ITEM_DEFAULTS['zx_add_constant' ] )

        if ENABLE_REGRESSION_LOOKING_BACK:
            self.recompute_regboundary.set( DEV_ITEM_DEFAULTS['recompute_regboundary' ] )


    def apply( self ):  # overrides parent class method

        ok_ = True

        if not ok_:
            MessageBox.showerror(
                error_msg_param[0],
                error_msg_param[1],
                parent=self,
                )
            self.applied    = None
            # setting this to None causes retry in the parent
            return

        if self.absorbance_has_been_changed and self.temp_absorbance is not None:
            # TODO: parent.serial_data does not exist!
            self.parent.serial_data.absorbance = self.temp_absorbance

        set_setting( 'data_exclusion',          self.data_exclusion.get() )
        set_setting( 'enable_affine_tran',      self.enable_affine_tran.get() )
        set_setting( 'decomp_from_separation',  self.decomp_from_separation.get() )
        if ENABLE_Rayleigh_Scatter_OPTION:
            set_setting( 'consider_scatter',    self.consider_scatter.get() )
        set_setting( 'almerge_analyzer',        self.almerge_analyzer.get() )
        set_setting( 'peak_mapping_only',       self.peak_mapping_only.get() )
        set_setting( 'lpm_variations',          self.lpm_variations.get() )
        set_setting( 'allow_angular_slope_in_mf', self.allow_angular_slope_in_mf.get() )

        revoke_excel = self.revoke_excel.get()
        set_setting( 'revoke_excel', revoke_excel )
        revoke_atsas = self.revoke_atsas.get()
        set_setting( 'revoke_atsas', revoke_atsas )
        if revoke_excel != self.revoke_excel_init or revoke_atsas != self.revoke_atsas_init:
            self.parent.check_environment()
        set_setting( 'revoke_cuda',             self.revoke_cuda.get() )

        set_setting( 'maintenance_mode',        self.maintenance_mode.get() )

        set_setting( 'fixed_guinier_start',     self.fixed_guinier_start.get() )
        set_setting( 'guinier_start_point',     self.guinier_start_point.get() )

        if ENABLE_DEPRECATED_OPTIONS:
            set_setting( 'zx_num_q_points',         self.regression_method.get() )
            set_setting( 'zx_build_method',         self.zx_build_method.get() )
            set_setting( 'zx_boundary_method',      self.zx_boundary_method.get() )
            set_setting( 'zx_boundary',             self.zx_boundary.get() )

        if ENABLE_Mapping_Animation:
            set_setting( 'enable_mapping_anim',     self.enable_mapping_anim.get() )

        set_setting( 'allow_rank_variation',    self.allow_rank_variation.get() )
        set_setting( 'enable_conctype_change',      self.enable_conctype_change.get() )
        set_setting( 'extended_conc_dep',       self.extended_conc_dep.get() )
        if False:
            set_setting( 'use_xray_conc',           self.use_xray_conc.get() )
            set_setting( 'apply_backsub',           self.apply_backsub.get() )

        set_dev_setting( 'enable_xb_save',      self.enable_xb_save.get() )
        set_dev_setting( 'xb_folder',           self.xb_folder.get() )
        set_dev_setting( 'base_file_postfix',   self.base_file_postfix.get() )

        if False:
            if ENABLE_XRAY_PROP_CONC_SMOOTHING:
                set_dev_setting( 'smoothed_xray_conc',  self.smoothed_xray_conc.get() )

        if ENABLE_REGRESSION_LOOKING_BACK:
            set_dev_setting( 'recompute_regboundary',   self.recompute_regboundary.get() )

        self.parent.update_plot_button_state()

        self.applied    = True

    def fixed_guinier_start_tracer( self, *args ):
        cb_value = self.fixed_guinier_start.get()
        state = Tk.NORMAL if cb_value == 1 else Tk.DISABLED
        for w in self.guinier_start_dependents:
            w.config( state=state )

    def guinier_plot_dialog( self ):
        from GuinierStartSelector   import GuinierStartSelector
        dialog = GuinierStartSelector( self, self.serial_data, self.init_ranges  )
        dialog.show()

    def zx_boundary_method_tracer( self, *args ):
        method = self.zx_boundary_method.get()
        state   = Tk.DISABLED if method == 'NO' else Tk.NORMAL
        for rb in self.zx_build_method_buttons:
            rb.config( state=state )

        state   = Tk.NORMAL if method == 'FIXED' else Tk.DISABLED
        self.zx_boundary_entry.config( state=state )



"""
    Models.Stochastic.SdmShapeStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import seaborn as sns

def study():
    from molass_legacy.Models.Stochastic.SdmGuessSingle import guess_sdm_params
    from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.Models.RateTheory.RobustEDM import edm_impl

    sns.set_theme()

    t0 = 50
    u = 1
    a = 1.0
    # b = -4.0
    b = 1.0
    e = 0.3
    Dz = 0.2
    cinj = 1

    x = np.arange(500)

    cinj = 1
    Dz = 0.5
    timescale = 0.25

    with plt.Dp():
        fig, axes = plt.subplots(ncols=3, figsize=(18, 5))
        fig.suptitle("SDM-EDM Correspondence Study", fontsize=20)
        for ax, b, title in zip(axes, [2, 0.0001, -2], ["Tailing Peak", "Symmetic Peak", "Fronting Peak"]):
            ax.set_title(title, fontsize=16)
            y = edm_impl(x, t0, u, a, b, e, Dz, cinj)
            ax.plot(x, y, label="EDM, b=%g" % b)
            K, N, t0, N0, rho, me, mp, scale = guess_sdm_params(x, y, timescale=timescale)
            T = K/N
            n1 = N*(1 - rho)**me
            t1 = T*(1 - rho)**mp
            y_ = scale*dispersive_monopore_pdf(x, n1, t1, N0, t0, timescale=timescale)
            ax.plot(x, y_, ":", lw=3, label="SDM, $t_0$=%.0f" % t0)
            if t0 > 0:
                ax.axvline(t0, color="red", label="$t_0$")
            ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("..\\lib")
    study()


"""
    DecompUtils.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.DataStructure.PeakInfo import PeakInfo
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.Models.ElutionCurveModels import EGH, EGHA, EMG, EMGA
from molass_legacy.Decomposer.ElutionDecomposer import ElutionDecomposer
from molass_legacy.Decomposer.ElutionModelScaler import scale_decomposed_elutions
from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo
from UnifiedDecompResult import UnifiedDecompResult
import molass_legacy.KekLib.DebugPlot as plt

NORMALIZED_RESID_LIMIT  = 0.02      # < 0.0476 for Kosugi3a
ALLOW_ONE_SIDE_RESULT   = False     # not allowing seems safe judging from 20230303/HasA
                                    # alternatively, consider making ElutionModelScaler more robust
ADD_PEAK_SCI_LIMIT = 70

def get_elm_info(e_elemnts, opt_recs):
    elm_recs = []
    for k, sel in enumerate(e_elemnts):
        if sel == 1:
            fnc = opt_recs[k][1]
            elm_recs.append( [ k, fnc ] )
    return elm_recs

def make_range_info_impl(opt_recs, control_info, specpanel_list=None, debug=False, parent=None, logger=None, no_elm_recs=False):
    select_matrix   = control_info.select_matrix
    editor_ranges   = control_info.editor_ranges
    top_x_list      = control_info.top_x_list

    if logger is not None:
        specpanel_list_info  = "" if specpanel_list is None else ", len(specpanel_list)=%d" % len(specpanel_list)
        logger.info('make_range_info_impl: editor_ranges=%s%s', str(editor_ranges), specpanel_list_info)

    """
        For debugging, see
            DecompEditorFrame.make_range_info
    """
    if debug:
        print('debug start')
        plt.push()
        fig = plt.figure()
        ax  = fig.gca()
        ax.set_title( 'make_range_info_impl debug' )
        x   = parent.mapper.x_curve.x

    ret_ranges = []
    for k, range_list in enumerate(editor_ranges):
        if specpanel_list is not None and len(specpanel_list) > 0:
            panel = specpanel_list[k]
            if len(range_list) == 0 or panel.ignore.get() == 1:
                continue

        top_x = top_x_list[k]
        if no_elm_recs:
            elm_recs = None
        else:
            e_elemnts = select_matrix[k]
            elm_recs = get_elm_info(e_elemnts, opt_recs)
            if debug:
                print([k], 'e_elemnts=', e_elemnts)
                for rec in elm_recs:
                    e   = rec[0]
                    fnc = rec[1]
                    ax.plot(x, fnc(x), label=str(e))
        ret_ranges.append( [ PeakInfo(k, top_x, elm_recs) ] + range_list  )

    if debug:
        print('debug show', 'ret_ranges=', ret_ranges)
        ax.legend()
        fig.tight_layout()
        plt.show()
        plt.pop()

    return ret_ranges

class CorrectedBaseline:
    def __init__(self, sd, mapper):
        x_curve = mapper.x_curve
        self.x_curve = x_curve
        self.x  = x_curve.x

        x_baseline = mapper.x_base + mapper.x_base_adjustment
        self.y  = mapper.x_vector - x_baseline
        self.max_y = self.y[x_curve.max_x]

        # TODO: confirm the necessity below
        if mapper.opt_params['xray_baseline_opt'] == 1:
            sd_copy = sd.get_exec_copy(mapper)      # to prevent given sd from being changed
            sd_copy.apply_baseline_correction( mapper.get_mapped_info() )
            sd = sd_copy

        self.data = sd.intensity_array
        self.sd = sd                    # added for VP-Analysis
        self.a_curve = mapper.a_curve   # added for VP-Analysis

def decompose_elution_impl(x_curve, x, y, model, dual_opt, hints_dict, d_curve, logger, debug):

    decomposer  = ElutionDecomposer(x_curve, x, y, model=model, d_curve=d_curve, retry_valley=True, deeply=True, hints_dict=hints_dict)

    if debug:
        print('len(decomposer.fit_recs)=', len(decomposer.fit_recs))

    if dual_opt:
        from DualOptimzer import get_dual_optimized_info, EmgaOptimzer, EghaOptimzer
        if model.get_name() == 'EMGA':
            optimizer_class = EmgaOptimzer
        else:
            optimizer_class = EghaOptimzer

        opt_recs = get_dual_optimized_info(optimizer_class, x_curve, x, y, decomposer.fit_recs, hints_dict, logger, debug=debug)
    else:
        opt_recs = decomposer.fit_recs

    return decomposer, opt_recs

def decompose_elution_xray_to_uv(corbase_info, mapper, model, a_curve, d_curve=None, dual_opt=True, hints_dict=None, logger=None, debug=False):
    x_curve = corbase_info.x_curve
    x       = corbase_info.x
    y       = corbase_info.y

    decomposer, opt_recs = decompose_elution_impl(x_curve, x, y, model, dual_opt, hints_dict, d_curve, logger, debug)
    # print('xray_to_uv(1): len(opt_recs)=', len(opt_recs))

    max_y = mapper.a_curve.max_y
    debug_info = None
    # debug_info = (x_curve, "decompose_elution_xray_to_uv")
    ret_uv = scale_decomposed_elutions(x, y, a_curve.y, max_y, opt_recs, model, logger, debug_info=debug_info)
    # print('xray_to_uv(2): len(opt_recs)=', len(opt_recs))

    opt_recs_uv = ret_uv.opt_recs
    model_name = model.get_name()

    if debug:
        debug_info = ("xray_to_uv", )
    else:
        debug_info = None

    return UnifiedDecompResult(
                    xray_to_uv=True,
                    x_curve=x_curve, x=x, y=y,
                    opt_recs=opt_recs,
                    max_y_xray = x_curve.max_y,
                    model_name=model_name,
                    decomposer=decomposer,
                    uv_y=a_curve.y,
                    opt_recs_uv=opt_recs_uv,
                    max_y_uv = max_y,
                    debug_info=debug_info,
                    )

def decompose_elution_uv_to_xray(corbase_info, mapper, model, a_curve, d_curve=None, dual_opt=True, hints_dict=None, logger=None, debug=False):
    x       = a_curve.x
    y       = a_curve.y
    decomposer, opt_recs_uv = decompose_elution_impl(a_curve, x, y, model, dual_opt, hints_dict, d_curve, logger, debug)
    # print('uv_to_xray(2): len(opt_recs_uv)=', len(opt_recs_uv))

    x_curve = corbase_info.x_curve
    xray_y  = corbase_info.y
    max_y   = corbase_info.max_y

    debug_info = None
    # debug_info = (a_curve, "decompose_elution_uv_to_xray")
    ret_xray = scale_decomposed_elutions(x, y, xray_y, max_y, opt_recs_uv, model, logger, debug_info=debug_info)
    # print('uv_to_xray(2): len(opt_recs_uv)=', len(opt_recs_uv))

    opt_recs = ret_xray.opt_recs
    model_name = model.get_name()

    if debug:
        debug_info = ("uv_to_xray", )
    else:
        debug_info = None

    return UnifiedDecompResult(
                    xray_to_uv=False,
                    x_curve=x_curve, x=x, y=xray_y,
                    opt_recs=opt_recs,
                    max_y_xray = max_y,
                    model_name=model_name,
                    decomposer=decomposer,
                    uv_y=a_curve.y,
                    opt_recs_uv=opt_recs_uv,
                    max_y_uv = a_curve.max_y,
                    debug_info=debug_info,
                    )

def decompose_elution_better(corbase_info, mapper, model=EMGA(), dual_opt=True, hints_dict=None, logger=None, return_both=False, debug=False):
    assert logger is not None

    mapped_uv_vector = mapper.make_uniformly_scaled_vector(scale=1)
    x_curve = corbase_info.x_curve
    a_curve = ElutionCurve(mapped_uv_vector)
    d_curve = None

    peaks = x_curve.get_emg_peaks()
    if len(peaks) == 1:
        min_sci = mapper.get_min_sci()
        if min_sci < ADD_PEAK_SCI_LIMIT:
            scale = mapper.get_x2a_scale()
            try:
                d_curve = ElutionCurve(np.abs(x_curve.y - a_curve.y/scale))
            except ZeroDivisionError:
                # as in 20180617
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(logger, "decompose_elution_better: could not make d_curve")
                d_curve = None

    if debug:
        from molass_legacy.Elution.CurveUtils import simple_plot
        scale = mapper.get_x2a_scale()
        m_curve = ElutionCurve(mapped_uv_vector/scale)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("decompose_elution_better entry")
            for c in [x_curve, m_curve, d_curve]:
                if c is not None:
                    simple_plot(ax, c)
            fig.tight_layout()
            plt.show()

    final_result = None
    x2u_result = decompose_elution_xray_to_uv(corbase_info, mapper, model, a_curve, d_curve, dual_opt, hints_dict, logger, debug)
    nresid1 = x2u_result.nresid_uv
    if ALLOW_ONE_SIDE_RESULT:
        if nresid1 < NORMALIZED_RESID_LIMIT and not return_both:
            logger.info('decomposer result has been adopted without comparison: %.3g' % (nresid1))
            final_result = x2u_result

    if final_result is None:
        try:
            u2x_result = decompose_elution_uv_to_xray(corbase_info, mapper, model, a_curve, d_curve, dual_opt, hints_dict, logger, debug)
        except AssertionError:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(logger, "decompose_elution_uv_to_xray failed: ")
            u2x_result = None

        if return_both:
            return x2u_result, u2x_result

        if u2x_result is None:
            final_result = x2u_result
        else:
            nresid2 = u2x_result.nresid_uv
            if nresid2 < nresid1:
                logger.info('decomposer result has been replaced after residual comparison: %.3g, %.3g' % (nresid1, nresid2))
                final_result = u2x_result
            else:
                logger.info('decomposer result has been kept after residual comparison: %.3g, %.3g' % (nresid1, nresid2))
                final_result = x2u_result

    final_result.remove_unwanted_elements()

    if debug:
        from OptRecsUtils import debug_plot_opt_recs
        debug_plot_opt_recs(x_curve, final_result.opt_recs, title="decompose_elution_better result: %s" % final_result.model_name )

    return final_result

def debug_elution_plot(title, model, x, y, opt_recs):
    plt.push()
    fig = plt.figure()
    ax = fig.gca()
    ax.set_title(title + model.get_name())
    ax.plot(x, y)
    for k, rec in enumerate(opt_recs):
        func = rec[1]
        y_ = func(x)
        ax.plot(x, y_, ':', label='element %d' % (k+1))

    ax.legend()
    fig.tight_layout()
    plt.show()
    plt.pop()

def get_peak2elem_dict(opt_recs):
    ret_dict = {}
    for eno, opt_rec in enumerate(opt_recs):
        pno = opt_rec[0]
        if pno < 0:
            continue

        ret_dict[pno] = eno

    print('opt_recs=', opt_recs)
    print('ret_dict=', ret_dict)
    return ret_dict

def make_default_analysis_range_info(sd, mapper, logger, select_fix=None):
    """
        currently, this is used only in test_870_ExtrapolationSolver.py

        TODO: recovery, where to apply LPM
    """
    corbase_info = CorrectedBaseline(sd, mapper)
    ret = decompose_elution_better(corbase_info, mapper, logger=logger)
    max_y = mapper.x_curve.max_y
    x = mapper.x_curve.x
    ret_info = ret.get_range_edit_info(logger=logger)

    if select_fix is not None:
        if select_fix.num_ranges == 0:
            k = select_fix.k
            update_selection(0, ret.opt_recs, select_matrix, k)
            editor_ranges[k].clear()
        else:
            assert False
        print( 'select_matrix=', select_matrix )

    ret_ranges = make_range_info_impl(ret.opt_recs_uv, ret_info)
    analysis_range_info = AnalysisRangeInfo( ret_ranges )
    return analysis_range_info



"""
    SSDC.FolderEntryDialog.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry

class FolderEntryDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, title="Input Folders Entry", visible=False)
        self.parent = parent
        self.applied = False
        self.info_list = None
    
    def show(self):
        self._show()
    
    def body(self, body_frame):
        self.entry1 = FolderEntry(body_frame, width=70)
        self.entry1.grid(row=0, column=0, padx=5, pady=5)
        self.entry2 = FolderEntry(body_frame, width=70)
        self.entry2.grid(row=1, column=0, padx=5, pady=5)
        nc_frame = Tk.Frame(body_frame)
        nc_frame.grid(row=2, column=0, padx=5, pady=5)
        self.num_components = Tk.IntVar()
        self.num_components.set(4)
        label = Tk.Label(nc_frame, text="Number of Components:")
        label.grid(row=0, column=0)
        self.spinbox = Tk.Spinbox(nc_frame, textvariable=self.num_components,
                                  from_=1, to=7, increment=1,
                                  justify=Tk.CENTER, width=6)        
        self.spinbox.grid(row=0, column=1)

        # self.entry1.variable.set('D:/PyTools/Data/20210727/data01')
        # self.entry2.variable.set('D:/PyTools/Data/20210727/data02')

        # self.entry1.variable.set('D:/PyTools/Data/20220716/BSA_201')
        # self.entry2.variable.set('D:/PyTools/Data/20220716/BSA_202')

        # self.entry1.variable.set('D:/PyTools/Data/20220716/FER_OA_301')
        # self.entry2.variable.set('D:/PyTools/Data/20220716/FER_OA_302')

        # self.entry1.variable.set('D:/PyTools/Data/20230706/ALD_OA001')
        # self.entry2.variable.set('D:/PyTools/Data/20230706/ALD_OA002')

        self.entry1.variable.set('D:/PyTools/Data/20230706/BSA001')
        self.entry2.variable.set('D:/PyTools/Data/20230706/BSA002')

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Test", width=10, command=self.test)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def apply(self):
        self.applied = True

    def test(self, devel=True):
        info_list = self.get_info_list()
        num_components = self.num_components.get()
        if devel:
            from importlib import reload
            import SSDC.SsdcAnalysis
            reload(SSDC.SsdcAnalysis)
        from SSDC.SsdcAnalysis import SsdcAnalysis
        try:
            analysis = SsdcAnalysis(self.parent, info_list, num_components)
            analysis.show()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "Error in SsdcAnalysis: ", n=10)

    def get_folder_list(self):
        ret_list = []
        for entry in [self.entry1, self.entry2]:
            ret_list.append(entry.variable.get())
        print(ret_list)
        return ret_list
    
    def get_info_list(self):
        if self.info_list is None:
            from molass_legacy.Batch.LiteBatch import LiteBatch        
            info_list = []
            batch = LiteBatch()     # task: 
            for folder in self.get_folder_list():               
                lrf_src = batch.get_lrf_source(in_folder=folder)
                info_list.append((folder, lrf_src))
            self.info_list = info_list
        return self.info_list


# coding: utf-8
"""
    ExtrapolationDenssMenu.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk
from MenuButton import MenuButton

class ExtrapolationDenssMenu(Tk.Frame):
    def __init__(self, parent, dialog, data_list):
        Tk.Frame.__init__(self, parent)
        self.parent = parent
        self.dialog = dialog
        self.data_list = data_list
        self.denss_menu = MenuButton(self, "DENSS Menu", [
                            ("Run all in background", self.submit_all),
                            ("Show DENSS Manager", self.show_denss_manager),
                            ("Show empty DENSS GUI", self.show_denss_gui),
                            ("Electron Density Viewer", self.show_ed_viewer),
                            ("SAXS Simulator", self.show_saxs_simulator),
                            ])
        self.denss_menu.pack()

    def submit_all(self):
        print('submit_all')
        from molass_legacy.DENSS.DenssUtils import fit_data
        from molass_legacy.DENSS.DenssManager import get_list
        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        job_list = get_list()
        num_jobs = len(job_list)
        num_jobs_str = "no" if num_jobs == 0 else "already " + str(num_jobs)
        yn = MessageBox.askyesno( "Run All Comfirmation",
            "There are %s jobs submitted.\n" % num_jobs_str
            + "Are you sure to run these %d jobs?" % len(self.data_list), parent=self.dialog )
        if not yn:
            return

        from molass_legacy.DENSS.DenssManager import JobInfo
        from molass_legacy.DENSS.DenssManagerDialog import show_manager_dialog
        jobs = []
        for q, a, e, f in self.data_list:
            qc, ac, ec, dmax = fit_data(q, a, e)
            jobs.append(JobInfo('denss', q=qc, a=ac, e=ec, dmax=dmax, infile_name=f))
        show_manager_dialog(self.dialog, jobs)

    def show_denss_gui(self):
        print('show_denss_gui')
        from molass_legacy.DENSS.DenssGui import DenssGuiDialog
        dialog = DenssGuiDialog(self.dialog)
        dialog.show()

    def show_denss_manager(self):
        print('show_denss_manager')
        from molass_legacy.DENSS.DenssManagerDialog import show_manager_dialog
        show_manager_dialog(self.dialog)

    def show_ed_viewer(self):
        print('show_ed_viewer')
        from molass_legacy.Saxs.EdViewer import EdViewer
        from molass_legacy.KekLib.OurMatplotlib import reset_to_default_style
        viewer = EdViewer(self.dialog)
        viewer.show()
        reset_to_default_style()

    def show_saxs_simulator(self):
        print('show_saxs_simulator')
        from molass_legacy.Saxs.SaxsSimulator import SaxsSimulator
        from molass_legacy.KekLib.OurMatplotlib import reset_to_default_style
        simulator = SaxsSimulator(self.parent)
        simulator.show()
        reset_to_default_style()



"""

    TkTester.py

    Copyright (c) 2017-2023, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
from time                   import sleep
import threading
from pyautogui              import typewrite
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from OurMessageBox          import dialog_queue

DEBUG               = False
COUNTER_CYCLE       = 10000
SHORT_INTERVAL      = 0.1
LONG_INTERVAL       = 0.5
LOOP_MAXNUM_CYCLE   = 360   # i.e., waits for LONG_INTERVAL * LOOP_MAXNUM_CYCLE seconds
GET_OBJ_MAX_RETRY   = 60
ENTER               = [ 'enter' ]
ESC                 = [ 'esc' ]

def reply_ok():
    # print( current_box_goemetry() )
    typewrite( ENTER )

def reply_yes():
    typewrite( 'Y' )

def reply_no():
    typewrite( 'N' )

lock = threading.Lock()

def messagebox_shown():
    return not dialog_queue.empty()

counter = 0
return_dict     = {}
attr_index_re   = re.compile( r'^(\w+)\[(\d+)\](.*)$' )
rest_index_re   = re.compile( r'^\[(\d+)\]' )

class Attribute(object):
    def __init__( self, parent, name ):
        self.__parent__ = parent
        self.__key__    = name

    def __getattribute__(self, key):
        if key.find( '__' ) == 0:
            if DEBUG: print( '(1) key=', key )
            return object.__getattribute__( self, key )
        else:
            attr = '.'.join( [ self.__key__, key ] )
            if DEBUG: print( 'Attribute:', attr )
            return Attribute( self.__parent__, attr )

    def __getitem__( self, n ):
        attr = self.__key__ + ( '[%d]' % n )
        return Attribute( self.__parent__, attr )

    def __call__( self, *args, __wait__=True, **wk_args ):
        sleep( SHORT_INTERVAL )
        if DEBUG: print( 'Attribute.__call__:', self.__key__,  *args )
        global counter
        if counter >= COUNTER_CYCLE:
            counter = 0

        counter += 1

        app = self.__parent__.__app__
        key = self.__key__
        log = self.__parent__.__log__

        success = False

        def get_obj():

            for retry in range( LOOP_MAXNUM_CYCLE ):
                if DEBUG: print( '(1) retry=', retry )
                try:
                    # this lock is probably necessary
                    if DEBUG: print( '(1) before lock' )
                    with lock:
                        if DEBUG: print( '(1-1) after lock' )
                        obj = app
                        for attr in key.split( '.' ):
                            if DEBUG: print( 'attr=', attr )
                            m = attr_index_re.match( attr )
                            if m:
                                attr_   = m.group(1)
                                n       = int( m.group(2) )
                                obj_    = object.__getattribute__( obj, attr_ )
                                obj     = obj_[n]
                                rest_index  = m.group(3)
                                if rest_index is None:
                                    pass
                                else:
                                    obj_ref = [obj]
                                    def apply_index( mobj ):
                                        i = int( mobj.group(1) )
                                        obj_ref[0] = obj_ref[0][i]
                                        return ''
                                    while rest_index != '':
                                        if DEBUG: print( 'rest_index=', rest_index )
                                        rest_index = re.sub( rest_index_re, apply_index, rest_index )
                                    obj = obj_ref[0]
                            else:
                                obj = object.__getattribute__( obj, attr )
                    success = True
                    break
                except AttributeError as exc:
                    print( 'Attribute: exc=', exc )
                    self.exception = exc
                    sleep( LONG_INTERVAL )
                    continue

            if DEBUG: print( '(1-2) after lock: obj=', obj )

            return success, obj

        success, obj = get_obj()
        if not success:
            print( 'key=', key, 'retry=', retry )
            raise self.exception

        should_wait = __wait__ is True or callable(__wait__)

        def invoke_closure( obj, counter ):
            if DEBUG: print( 'invoke_closure: counter=', counter )
            retry_count = 0
            while True:
                try:
                    if should_wait:
                        ckey = counter
                    else:
                        return_dict[ counter ] = counter
                        ckey = -counter     # negative key values will remain until clearance
                                            # without interfering positive key referencing
                    # ret = obj( *args )
                    ret = obj( *args, **wk_args )
                    return_dict[ ckey ] = ret
                    if DEBUG:
                        print( '---------- invoke_closure success', retry_count)
                    break
                except Exception as exc:
                    # one of the cases that cause this exception is a too long delay in
                    # DLL loading under, e.g, the real-time scan by anti-virus software.
                    print( exc )
                    if log is not None:
                        log( str(exc) + '\n' )
                    n = str(exc).find('invalid command name')
                    # n = str(exc).find('unknown')
                    print( 'ERROR at ' + key, n )
                    if n >= 0:
                        app.update()
                        sleep(1)
                        retry_count += 1
                        if DEBUG:
                            print("---------- retry_count=", retry_count)
                        if retry_count < GET_OBJ_MAX_RETRY:
                            success, obj = get_obj()
                            continue
                        else:
                            raise exc
                    else:
                        raise exc

        # not sure whethrer this lock is necessary
        if DEBUG: print( '(2) before lock' )
        with lock:
            if DEBUG: print( '(2-1) after lock' )
            app.after( 100, lambda c=counter: invoke_closure( obj, c ) )
        if DEBUG: print( '(2-2) after lock' )

        sleep( 0.2 )    # to probably succeed at the first try

        # note that "not __wait" cases wait for the invoke_closure call
        ret = None
        for retry in range( LOOP_MAXNUM_CYCLE ):
            if DEBUG: print( '(2) retry=', retry )
            try:
                ret = return_dict[ counter ]
                del return_dict[ counter ]
                break
            except KeyError:
                if DEBUG: print( 'retry=', retry, 'counter=', counter )
                try:
                    app.state()
                except Exception as exc:
                    print( exc )
                    break

                if callable( __wait__ ):
                    if not __wait__( retry ):
                        break

                sleep( LONG_INTERVAL )  # long or short?
                continue

        if DEBUG: print( '__call__: return' )
        return ret

class AppAgent(object):
    def __init__( self, app, log_func=None ):
        self.__app__    = app
        self.__log__    = log_func

    def __getattribute__( self, key ):
        if key.find( '__' ) == 0:
            if DEBUG: print( '(2) key=', key )
            return object.__getattribute__( self, key )
        else:
            print( 'AppAgent:',  key )
            return Attribute( self, key )

    def __is_alive__( self ):
        ret = False
        try:
            self.__app__.state()
            ret = True
        except Exception as exc:
            print( exc )
            pass
        return ret

class TestClient:
    def __init__( self, app, script, log_func=None ):
        self.agent = AppAgent( app, log_func=log_func )
        self.thread = threading.Thread(
                        target=script,
                        name='TestClientThread',
                        args=[ self, self.agent ]
                        )
        self.thread.start()

    def quit(self):
        self.thread.join()

    def __del__(self):
        # seems to be called in the current self.thread
        print('__del__')
